import datetime
import itertools
import json
import os
import socket
from collections import namedtuple
from copy import deepcopy
from time import sleep
from typing import Any, Callable, Dict, Optional, Sequence, Tuple

from pystencils.runhelper import Database
from pystencils.utils import DotDict

ParameterDict = Dict[str, Any]
WeightFunction = Callable[[Dict], int]
FilterFunction = Callable[[ParameterDict], Optional[ParameterDict]]


class ParameterStudy:
    """Manages and runs multiple configurations locally or distributed and stores results in NoSQL database.

    To run a parameter study, define a run function that takes all parameters as keyword arguments and returns the
    results as a (possibly nested) dictionary. Then, define the parameter sets that this function should be run with.

    Examples:
          >>> import tempfile
          >>>
          >>> def dummy_run_function(p1, p2, p3, p4):
          ...     print("Run called with", p1, p2, p3, p4)
          ...     return { 'result1': p1 * p2, 'result2': p3 + p4 }
          >>>
          >>> with tempfile.TemporaryDirectory() as tmp_dir:
          ...     ps = ParameterStudy(dummy_run_function, database_connector=tmp_dir)
          ...     ps.add_run({'p1': 5, 'p2': 42, 'p3': 'abc', 'p4': 'def'})
          ...     ps.add_combinations( [('p1', [1, 2]),
          ...                           ('p3', ['x', 'y'])], constant_parameters={'p2': 5, 'p4': 'z' })
          ...     ps.run()
          ...     ps.run_scenarios_not_in_database()
          ...     ps.run_from_command_line(argv=['local'])  # alternative to run - exposes a command line interface if
          ...                                               # no argv is passed. Does not run anything here, because
          ...                                               # configuration already in database are skipped
          Run called with 2 5 y z
          Run called with 2 5 x z
          Run called with 1 5 y z
          Run called with 1 5 x z
          Run called with 5 42 abc def

    Above example runs all parameter combinations locally and stores the returned result in the NoSQL database.
    It is also possible to distribute the runs to multiple processes, by starting a server on one machine and multiple
    executing runners on other machines. The server distributes configurations to the runners, collects their results
    to stores the results in the database.
    """

    Run = namedtuple("Run", ['parameter_dict', 'weight'])

    def __init__(self, run_function: Callable[..., Dict], runs: Sequence = (),
                 database_connector: str = './db') -> None:
        self.runs = list(runs)
        self.run_function = run_function
        self.db = Database(database_connector)

    def add_run(self, parameter_dict: ParameterDict, weight: int = 1) -> None:
        """Schedule a dictionary of parameters to run in this parameter study.

        Args:
            parameter_dict: used as keyword arguments to the run function.
            weight: weight of the run configuration which should be  proportional to runtime of this case,
                    used for progress display and distribution to processes.
        """
        self.runs.append(self.Run(parameter_dict, weight))

    def add_combinations(self, degrees_of_freedom: Sequence[Tuple[str, Sequence[Any]]],
                         constant_parameters: Optional[ParameterDict] = None,
                         filter_function: Optional[FilterFunction] = None,
                         runtime_weight_function: Optional[WeightFunction] = None) -> None:
        """Add all possible combinations of given parameters as runs.

        This is a convenience function to simulate all possible parameter combinations of a scenario.
        Configurations can be filtered and weighted by passing filter- and weighting functions.

        Args:
            degrees_of_freedom: defines for each parameter the possible values it can take on
            constant_parameters: parameter dict, for parameters that should not be changed
            filter_function: optional function that receives a parameter dict and returns the potentially modified dict
                             or None if this combination should not be added.
            runtime_weight_function: function mapping a parameter dict to the runtime weight (see weight at add_runs)

        Examples:
             degrees_of_freedom = [('p1', [1,2]),
                                   ('p2', ['a', 'b'])]
             is equivalent to calling add_run four times, with all possible parameter combinations.
        """
        parameter_names = [e[0] for e in degrees_of_freedom]
        parameter_values = [e[1] for e in degrees_of_freedom]

        default_params_dict = {} if constant_parameters is None else constant_parameters
        for value_tuple in itertools.product(*parameter_values):
            params_dict = deepcopy(default_params_dict)
            params_dict.update({name: value for name, value in zip(parameter_names, value_tuple)})
            params = DotDict(params_dict)
            if filter_function:
                params = filter_function(params)
                if params is None:
                    continue
            weight = 1 if not runtime_weight_function else runtime_weight_function(params)
            self.add_run(params, weight)

    def run(self, process: int = 0, num_processes: int = 1, parameter_update: Optional[ParameterDict] = None) -> None:
        """Runs all added configurations.

        Args:
            process: configurations are split into num_processes chunks according to weights and only the
                     process'th chunk is run. To run all, use process=0 and num_processes=1
            num_processes: see above
            parameter_update: Extend/override all configurations with this dictionary.
        """
        parameter_update = {} if parameter_update is None else parameter_update
        own_runs = self._distribute_runs(self.runs, process, num_processes)
        for run in own_runs:
            parameter_dict = run.parameter_dict.copy()
            parameter_dict.update(parameter_update)
            result = self.run_function(**parameter_dict)

            self.db.save(run.parameter_dict, result, None, changed_params=parameter_update)

    def run_scenarios_not_in_database(self, parameter_update: Optional[ParameterDict] = None) -> None:
        """Same as run method, but runs only configuration for which no result is in the database yet."""
        parameter_update = {} if parameter_update is None else parameter_update
        filtered_runs = self._filter_already_simulated(self.runs)
        for run in filtered_runs:
            parameter_dict = run.parameter_dict.copy()
            parameter_dict.update(parameter_update)
            result = self.run_function(**parameter_dict)

            self.db.save(run.parameter_dict, result, changed_params=parameter_update)

    def run_server(self, ip: str = "0.0.0.0", port: int = 8342):
        """Runs server to supply runner clients with scenarios to simulate and collect results from them.
        Skips scenarios that are already in the database."""
        from http.server import BaseHTTPRequestHandler, HTTPServer
        filtered_runs = self._filter_already_simulated(self.runs)

        if not filtered_runs:
            print("No Scenarios to simulate")
            return

        class ParameterStudyServer(BaseHTTPRequestHandler):
            parameterStudy = self
            all_runs = filtered_runs
            runs = filtered_runs.copy()
            currently_running = {}
            finished_runs = []

            def next_scenario(self, received_json_data):
                client_name = received_json_data['client_name']
                if len(self.runs) > 0:
                    run_status = "%d/%d" % (len(self.finished_runs), len(self.all_runs))
                    work_status = "%d/%d" % (sum(r.weight for r in self.finished_runs),
                                             sum(r.weight for r in self.all_runs))
                    format_args = {
                        'remaining': len(self.runs),
                        'time': datetime.datetime.now().strftime("%H:%M:%S"),
                        'client_name': client_name,
                        'run_status': run_status,
                        'work_status': work_status,
                    }

                    scenario = self.runs.pop(0)
                    print(" {time} {client_name} fetched scenario. Scenarios: {run_status}, Work: {work_status}"
                          .format(**format_args))
                    self.currently_running[client_name] = scenario
                    return {'status': 'ok', 'params': scenario.parameter_dict}
                else:
                    return {'status': 'finished'}

            def result(self, received_json_data):
                client_name = received_json_data['client_name']
                run = self.currently_running[client_name]
                self.finished_runs.append(run)
                del self.currently_running[client_name]
                d = received_json_data

                def hash_dict(dictionary):
                    import hashlib
                    return hashlib.sha1(json.dumps(dictionary, sort_keys=True).encode()).hexdigest()

                assert hash_dict(d['params']) == hash_dict(run.parameter_dict), \
                    str(d['params']) + "is not equal to " + str(run.parameter_dict)
                self.parameterStudy.db.save(run.parameter_dict,
                                            result=d['result'], env=d['env'], changed_params=d['changed_params'])
                return {}

            # noinspection PyPep8Naming
            def do_POST(self) -> None:
                mapping = {'/next_scenario': self.next_scenario,
                           '/result': self.result}
                if self.path in mapping.keys():
                    data = self._read_contents()
                    self.send_response(200)
                    self.send_header("Content-type", "application/json")
                    self.end_headers()
                    json_data = json.loads(data)
                    response = mapping[self.path](json_data)
                    self.wfile.write(json.dumps(response).encode())
                else:
                    self.send_response(400)

            # noinspection PyPep8Naming
            def do_GET(self):
                return self.do_POST()

            def _read_contents(self):
                return self.rfile.read(int(self.headers['Content-Length'])).decode()

            def log_message(self, fmt, *args):
                return

        print(f"Listening to connections on {ip}:{port}. Scenarios to simulate: {len(filtered_runs)}")
        server = HTTPServer((ip, port), ParameterStudyServer)
        while len(ParameterStudyServer.currently_running) > 0 or len(ParameterStudyServer.runs) > 0:
            server.handle_request()
        server.handle_request()

    def run_client(self, client_name: str = "{hostname}_{pid}", server: str = 'localhost', port: int = 8342,
                   parameter_update: Optional[ParameterDict] = None, max_time=None) -> None:
        """Start runner client that retrieves configuration from server, runs it and reports results back to server.

        Args:
            client_name: name of the client. Has to be unique for each client.
                         Placeholders {hostname} and {pid} can be used to generate unique name.
            server: url to server
            port: port as specified in run_server
            parameter_update: Used to override/extend parameters received from the server.
                              Typical use cases is to set optimization or GPU parameters for some clients to make
                              some clients simulate on CPU, others on GPU
            max_time: maximum runtime in seconds: the client runs scenario after scenario, but starts only a new
                      scenario if not more than max_time seconds have passed since this function was called.
                      So the time given here should be the total maximum runtime minus a typical runtime for one setup
        """
        from urllib.request import urlopen
        from urllib.error import URLError
        import time
        parameter_update = {} if parameter_update is None else parameter_update
        url = f"http://{server}:{port}"
        client_name = client_name.format(hostname=socket.gethostname(), pid=os.getpid())
        start_time = time.time()
        while True:
            try:
                if max_time is not None and (time.time() - start_time) > max_time:
                    print("Stopping client - maximum time reached")
                    break
                http_response = urlopen(url + "/next_scenario",
                                        data=json.dumps({'client_name': client_name}).encode())
                scenario = json.loads(http_response.read().decode())
                if scenario['status'] != 'ok':
                    break
                original_params = scenario['params'].copy()
                scenario['params'].update(parameter_update)
                result = self.run_function(**scenario['params'])

                answer = {'params': original_params,
                          'changed_params': parameter_update,
                          'result': result,
                          'env': Database.get_environment(),
                          'client_name': client_name}
                urlopen(url + '/result', data=json.dumps(answer).encode())
            except URLError:
                print(f"Cannot connect to server {url}  retrying in 5 seconds...")
                sleep(5)

    def run_from_command_line(self, argv: Optional[Sequence[str]] = None) -> None:
        """Exposes interface to command line with possibility to run directly or distributed via server/client."""
        from argparse import ArgumentParser

        def server(a):
            if a.database:
                self.db = Database(a.database)
            self.run_server(a.host, a.port)

        def client(a):
            self.run_client(a.client_name, a.host, a.port, json.loads(a.parameter_override), a.max_time)

        def local(a):
            if a.database:
                self.db = Database(a.database)
            self.run_scenarios_not_in_database(json.loads(a.parameter_override))

        parser = ArgumentParser()
        subparsers = parser.add_subparsers()

        local_parser = subparsers.add_parser('local', aliases=['l'],
                                             help="Run scenarios locally which are not yet in database", )
        local_parser.add_argument("-d", "--database", type=str, default="")
        local_parser.add_argument("-P", "--parameter_override", type=str, default="{}",
                                  help="JSON: the parameter dictionary is updated with these parameters. Use this to "
                                       "set host specific options like GPU call parameters. Enclose in \" ")
        local_parser.set_defaults(func=local)

        server_parser = subparsers.add_parser('server', aliases=['s'],
                                              help="Runs server to distribute different scenarios to workers", )
        server_parser.add_argument("-p", "--port", type=int, default=8342, help="Port to listen on")
        server_parser.add_argument("-H", "--host", type=str, default="0.0.0.0", help="IP/Hostname to listen on")
        server_parser.add_argument("-d", "--database", type=str, default="")
        server_parser.set_defaults(func=server)

        client_parser = subparsers.add_parser('client', aliases=['c'],
                                              help="Runs a worker client connection to scenario distribution server")
        client_parser.add_argument("-p", "--port", type=int, default=8342, help="Port to connect to")
        client_parser.add_argument("-H", "--host", type=str, default="localhost", help="Host or IP to connect to")
        client_parser.add_argument("-n", "--client_name", type=str, default="{hostname}_{pid}",
                                   help="Unique client name, you can use {hostname} and {pid} as placeholder")
        client_parser.add_argument("-P", "--parameter_override", type=str, default="{}",
                                   help="JSON: the parameter dictionary is updated with these parameters. Use this to "
                                        "set host specific options like GPU call parameters. Enclose in \" ")
        client_parser.add_argument("-t", "--max_time", type=int, default=None,
                                   help="If more than this time in seconds has passed, "
                                        "the client stops running scenarios.")
        client_parser.set_defaults(func=client)

        args = parser.parse_args(argv)
        if not len(vars(args)):
            parser.print_help()
        else:
            args.func(args)

    def _filter_already_simulated(self, all_runs):
        """Removes all runs from the given list, that are already in the database"""
        already_simulated = {json.dumps(e.params) for e in self.db.filter({})}
        return [r for r in all_runs if json.dumps(r.parameter_dict) not in already_simulated]

    @staticmethod
    def _distribute_runs(all_runs, process, num_processes):
        """Partitions runs by their weights into num_processes chunks and returns the process's chunk."""
        sorted_runs = sorted(all_runs, key=lambda e: e.weight, reverse=True)
        result = sorted_runs[process::num_processes]
        result.reverse()  # start with faster scenarios
        return result



import pytest
import numpy as np

import pystencils as ps
from pystencils import Field, Assignment, create_kernel
from pystencils.bit_masks import flag_cond


@pytest.mark.parametrize('mask_type', [np.uint8, np.uint16, np.uint32, np.uint64])
def test_flag_condition(mask_type):
    f_arr = np.zeros((2, 2, 2), dtype=np.float64)
    mask_arr = np.zeros((2, 2), dtype=mask_type)

    mask_arr[0, 1] = (1 << 3)
    mask_arr[1, 0] = (1 << 5)
    mask_arr[1, 1] = (1 << 3) + (1 << 5)

    f = Field.create_from_numpy_array('f', f_arr, index_dimensions=1)
    mask = Field.create_from_numpy_array('mask', mask_arr)

    v1 = 42.3
    v2 = 39.7
    v3 = 119

    assignments = [
        Assignment(f(0), flag_cond(3, mask(0), v1)),
        Assignment(f(1), flag_cond(5, mask(0), v2, v3))
    ]

    kernel = create_kernel(assignments).compile()
    kernel(f=f_arr, mask=mask_arr)
    code = ps.get_code_str(kernel)
    assert '119.0' in code

    reference = np.zeros((2, 2, 2), dtype=np.float64)
    reference[0, 1, 0] = v1
    reference[1, 1, 0] = v1

    reference[0, 0, 1] = v3
    reference[0, 1, 1] = v3

    reference[1, 0, 1] = v2
    reference[1, 1, 1] = v2

    np.testing.assert_array_equal(f_arr, reference)



import sympy as sp
from pystencils import fields
from pystencils.fd import Diff, diff, collect_diffs
from pystencils.fd.derivative import replace_generic_laplacian


def test_fs():
    f = sp.Symbol("f", commutative=False)

    a = Diff(Diff(Diff(f, 1), 0), 0)
    assert a.is_commutative is False
    print(str(a))

    assert diff(f) == f

    x, y = sp.symbols("x, y")
    collected_terms = collect_diffs(diff(x, 0, 0))
    assert collected_terms == Diff(Diff(x, 0, -1), 0, -1)

    src = fields("src : double[2D]")
    expr = sp.Add(Diff(Diff(src[0, 0])), 10)
    expected = Diff(Diff(src[0, 0], 0, -1), 0, -1) + Diff(Diff(src[0, 0], 1, -1), 1, -1) + 10
    result = replace_generic_laplacian(expr, 3)
    assert result == expected


import os
import runpy
import sys
import tempfile
import warnings

import nbformat
import pytest
from nbconvert import PythonExporter

from pystencils.boundaries.createindexlistcython import *  # NOQA
# Trigger config file reading / creation once - to avoid race conditions when multiple instances are creating it
# at the same time
from pystencils.cpu import cpujit

# trigger cython imports - there seems to be a problem when multiple processes try to compile the same cython file
# at the same time
try:
    import pyximport
    pyximport.install(language_level=3)
except ImportError:
    pass


SCRIPT_FOLDER = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, os.path.abspath('pystencils'))

# the Ubuntu pipeline uses an older version of pytest which uses deprecated functionality.
# This leads to many warinings in the test and coverage pipeline.
pytest_numeric_version = [int(x, 10) for x in pytest.__version__.split('.')]
pytest_numeric_version.reverse()
pytest_version = sum(x * (100 ** i) for i, x in enumerate(pytest_numeric_version))


def add_path_to_ignore(path):
    if not os.path.exists(path):
        return
    global collect_ignore
    collect_ignore += [os.path.join(SCRIPT_FOLDER, path, f) for f in os.listdir(os.path.join(SCRIPT_FOLDER, path))]


collect_ignore = [os.path.join(SCRIPT_FOLDER, "doc", "conf.py"),
                  os.path.join(SCRIPT_FOLDER, "pystencils", "opencl", "opencl.autoinit")]
add_path_to_ignore('pystencils_tests/benchmark')
add_path_to_ignore('_local_tmp')


try:
    import pycuda
except ImportError:
    collect_ignore += [os.path.join(SCRIPT_FOLDER, "pystencils_tests/test_cudagpu.py")]
    add_path_to_ignore('pystencils/gpucuda')

try:
    import waLBerla
except ImportError:
    collect_ignore += [os.path.join(SCRIPT_FOLDER, "pystencils_tests/test_aligned_array.py"),
                       os.path.join(SCRIPT_FOLDER, "pystencils_tests/test_datahandling_parallel.py"),
                       os.path.join(SCRIPT_FOLDER, "doc/notebooks/03_tutorial_datahandling.ipynb"),
                       os.path.join(SCRIPT_FOLDER, "pystencils/datahandling/parallel_datahandling.py"),
                       os.path.join(SCRIPT_FOLDER, "pystencils_tests/test_small_block_benchmark.ipynb")]

try:
    import blitzdb
except ImportError:
    add_path_to_ignore('pystencils/runhelper')
    collect_ignore += [os.path.join(SCRIPT_FOLDER, "pystencils_tests/test_parameterstudy.py")]

try:
    import islpy
except ImportError:
    collect_ignore += [os.path.join(SCRIPT_FOLDER, "pystencils/integer_set_analysis.py")]

try:
    import graphviz
except ImportError:
    collect_ignore += [os.path.join(SCRIPT_FOLDER, "pystencils/backends/dot.py")]
    collect_ignore += [os.path.join(SCRIPT_FOLDER, "doc/notebooks/01_tutorial_getting_started.ipynb")]

try:
    import pyevtk
except ImportError:
    collect_ignore += [os.path.join(SCRIPT_FOLDER, "pystencils/datahandling/vtk.py")]

collect_ignore += [os.path.join(SCRIPT_FOLDER, 'setup.py')]

for root, sub_dirs, files in os.walk('.'):
    for f in files:
        if f.endswith(".ipynb") and not any(f.startswith(k) for k in ['demo', 'tutorial', 'test', 'doc']):
            collect_ignore.append(f)


class IPythonMockup:
    def run_line_magic(self, *args, **kwargs):
        pass

    def run_cell_magic(self, *args, **kwargs):
        pass

    def magic(self, *args, **kwargs):
        pass

    def __bool__(self):
        return False


class IPyNbTest(pytest.Item):
    def __init__(self, name, parent, code):
        super(IPyNbTest, self).__init__(name, parent)
        self.code = code
        self.add_marker('notebook')

    @pytest.mark.filterwarnings("ignore:IPython.core.inputsplitter is deprecated")
    def runtest(self):
        global_dict = {'get_ipython': lambda: IPythonMockup(),
                       'is_test_run': True}

        # disable matplotlib output
        exec("import matplotlib.pyplot as p; "
             "p.switch_backend('Template')", global_dict)

        # in notebooks there is an implicit plt.show() - if this is not called a warning is shown when the next
        # plot is created. This warning is suppressed here
        exec("import warnings;"
             "warnings.filterwarnings('ignore', 'Adding an axes using the same arguments as a previous.*')",
             global_dict)
        with tempfile.NamedTemporaryFile() as f:
            f.write(self.code.encode())
            f.flush()
            runpy.run_path(f.name, init_globals=global_dict, run_name=self.name)


class IPyNbFile(pytest.File):
    def collect(self):
        exporter = PythonExporter()
        exporter.exclude_markdown = True
        exporter.exclude_input_prompt = True

        notebook_contents = self.fspath.open(encoding='utf-8')

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "IPython.core.inputsplitter is deprecated")
            notebook = nbformat.read(notebook_contents, 4)
            code, _ = exporter.from_notebook_node(notebook)
        if pytest_version >= 50403:
            yield IPyNbTest.from_parent(name=self.name, parent=self, code=code)
        else:
            yield IPyNbTest(self.name, self, code)

    def teardown(self):
        pass


def pytest_collect_file(path, parent):
    glob_exprs = ["*demo*.ipynb", "*tutorial*.ipynb", "test_*.ipynb"]
    if any(path.fnmatch(g) for g in glob_exprs):
        if pytest_version >= 50403:
            return IPyNbFile.from_parent(fspath=path, parent=parent)
        else:
            return IPyNbFile(path, parent)



import pystencils as ps

from pystencils.astnodes import Block, Conditional, SympyAssignment


def test_dot_print():
    src, dst = ps.fields("src, dst: double[2D]", layout='c')

    true_block = Block([SympyAssignment(dst[0, 0], src[-1, 0])])
    false_block = Block([SympyAssignment(dst[0, 0], src[1, 0])])
    ur = [true_block, Conditional(dst.center() > 0.0, true_block, false_block)]

    ast = ps.create_kernel(ur)



import os
from tempfile import TemporaryDirectory

import numpy as np
import pytest

import pystencils
from pystencils import Assignment, create_kernel
from pystencils.boundaries import BoundaryHandling, Dirichlet, Neumann, add_neumann_boundary
from pystencils.datahandling import SerialDataHandling
from pystencils.enums import Target
from pystencils.slicing import slice_from_direction
from pystencils.timeloop import TimeLoop


def test_kernel_vs_copy_boundary():
    dh = SerialDataHandling(domain_size=(7, 7))
    src = dh.add_array('src')
    dst_builtin = dh.add_array_like('dst_builtin', 'src')
    dst_python_copy = dh.add_array_like('dst_python_copy', 'src')
    dst_handling = dh.add_array_like('dst_handling', 'src')

    src_arr = np.arange(dh.shape[0] * dh.shape[1]).reshape(dh.shape)

    def reset_src():
        for block in dh.iterate(ghost_layers=True, inner_ghost_layers=True):
            np.copyto(block['src'], np.random.rand(*block.shape))

        for block in dh.iterate(ghost_layers=False, inner_ghost_layers=True):
            np.copyto(block['src'], src_arr)

    for b in dh.iterate(ghost_layers=False, inner_ghost_layers=True):
        np.copyto(b['dst_builtin'], 42)
        np.copyto(b['dst_python_copy'], 43)
        np.copyto(b['dst_handling'], 44)

    flags = dh.add_array('flags', dtype=np.uint8)
    dh.fill(flags.name, 0)
    borders = ['N', 'S', 'E', 'W']
    for d in borders:
        dh.fill(flags.name, 1, slice_obj=slice_from_direction(d, dim=2), ghost_layers=True, inner_ghost_layers=True)

    rhs = sum(src.neighbors([(1, 0), (-1, 0), (0, 1), (0, -1)]))

    simple_kernel = create_kernel([Assignment(dst_python_copy.center, rhs)]).compile()
    kernel_handling = create_kernel([Assignment(dst_handling.center, rhs)]).compile()

    assignments_with_boundary = add_neumann_boundary([Assignment(dst_builtin.center, rhs)],
                                                     fields=[src], flag_field=flags, boundary_flag=1)
    kernel_with_boundary = create_kernel(assignments_with_boundary).compile()

    # ------ Method 1: Built-in boundary
    reset_src()
    dh.run_kernel(kernel_with_boundary)

    # ------ Method 2: Using python to copy out the values (reference)
    reset_src()
    for b in dh.iterate():
        arr = b['src']
        arr[:, 0] = arr[:, 1]
        arr[:, -1] = arr[:, -2]
        arr[0, :] = arr[1, :]
        arr[-1, :] = arr[-2, :]
    dh.run_kernel(simple_kernel)

    # ------ Method 3: Using boundary handling to copy out the values
    reset_src()
    boundary_stencil = [(1, 0), (-1, 0), (0, 1), (0, -1)]
    boundary_handling = BoundaryHandling(dh, src.name, boundary_stencil)
    neumann = Neumann()
    assert neumann.name == 'Neumann'
    neumann.name = "wall"
    assert neumann.name == 'wall'
    assert neumann.additional_data_init_callback is None
    assert len(neumann.additional_data) == 0

    for d in ('N', 'S', 'W', 'E'):
        boundary_handling.set_boundary(neumann, slice_from_direction(d, dim=2))
    boundary_handling()
    dh.run_kernel(kernel_handling)

    python_copy_result = dh.gather_array('dst_python_copy')
    builtin_result = dh.gather_array('dst_builtin')
    handling_result = dh.gather_array('dst_handling')

    np.testing.assert_almost_equal(python_copy_result, builtin_result)
    np.testing.assert_almost_equal(python_copy_result, handling_result)

    with TemporaryDirectory() as tmp_dir:
        pytest.importorskip('pyevtk')
        boundary_handling.geometry_to_vtk(file_name=os.path.join(tmp_dir, 'test_output1'), ghost_layers=False)
        boundary_handling.geometry_to_vtk(file_name=os.path.join(tmp_dir, 'test_output2'), ghost_layers=True)

        boundaries = list(boundary_handling._boundary_object_to_boundary_info.keys()) + ['domain']
        boundary_handling.geometry_to_vtk(file_name=os.path.join(tmp_dir, 'test_output3'),
                                          boundaries=boundaries[0], ghost_layers=False)


def test_boundary_gpu():
    pytest.importorskip('pycuda')
    dh = SerialDataHandling(domain_size=(7, 7), default_target=Target.GPU)
    src = dh.add_array('src')
    dh.fill("src", 0.0, ghost_layers=True)
    dh.fill("src", 1.0, ghost_layers=False)
    src_cpu = dh.add_array('src_cpu', gpu=False)
    dh.fill("src_cpu", 0.0, ghost_layers=True)
    dh.fill("src_cpu", 1.0, ghost_layers=False)

    boundary_stencil = [(1, 0), (-1, 0), (0, 1), (0, -1)]
    boundary_handling_cpu = BoundaryHandling(dh, src_cpu.name, boundary_stencil,
                                             name="boundary_handling_cpu", target=Target.CPU)

    boundary_handling = BoundaryHandling(dh, src.name, boundary_stencil,
                                         name="boundary_handling_gpu", target=Target.GPU)

    neumann = Neumann()
    for d in ('N', 'S', 'W', 'E'):
        boundary_handling.set_boundary(neumann, slice_from_direction(d, dim=2))
        boundary_handling_cpu.set_boundary(neumann, slice_from_direction(d, dim=2))

    boundary_handling.prepare()
    boundary_handling_cpu.prepare()

    boundary_handling_cpu()

    dh.all_to_gpu()
    boundary_handling()
    dh.all_to_cpu()
    np.testing.assert_almost_equal(dh.cpu_arrays["src_cpu"], dh.cpu_arrays["src"])


def test_boundary_utility():
    dh = SerialDataHandling(domain_size=(7, 7))
    src = dh.add_array('src')
    dh.fill("src", 0.0, ghost_layers=True)

    boundary_stencil = [(1, 0), (-1, 0), (0, 1), (0, -1)]

    boundary_handling = BoundaryHandling(dh, src.name, boundary_stencil,
                                         name="boundary_handling", target=Target.CPU)

    neumann = Neumann()
    dirichlet = Dirichlet(2)
    for d in ('N', 'S', 'W', 'E'):
        boundary_handling.set_boundary(neumann, slice_from_direction(d, dim=2))

    boundary_handling.set_boundary(neumann, (slice(2, 4, None), slice(2, 4, None)))

    boundary_handling.prepare()

    assert boundary_handling.get_flag(boundary_handling.boundary_objects[0]) == 2
    assert boundary_handling.shape == dh.shape
    assert boundary_handling.flag_array_name == 'boundary_handlingFlags'
    mask_neumann = boundary_handling.get_mask((slice(0, 7), slice(0, 7)), boundary_handling.boundary_objects[0])
    np.testing.assert_almost_equal(mask_neumann[1:3, 1:3], 2)

    mask_domain = boundary_handling.get_mask((slice(0, 7), slice(0, 7)), "domain")
    assert np.sum(mask_domain) == 7 ** 2 - 4

    def set_sphere(x, y):
        mid = (4, 4)
        radius = 2
        return (x - mid[0]) ** 2 + (y - mid[1]) ** 2 < radius ** 2

    boundary_handling.set_boundary(dirichlet, mask_callback=set_sphere, force_flag_value=4)
    mask_dirichlet = boundary_handling.get_mask((slice(0, 7), slice(0, 7)), boundary_handling.boundary_objects[1])
    assert np.sum(mask_dirichlet) == 48

    assert boundary_handling.set_boundary("domain") == 1

    assert boundary_handling.set_boundary(dirichlet, mask_callback=set_sphere, force_flag_value=8, replace=False) == 4
    assert boundary_handling.set_boundary(dirichlet, force_flag_value=16, replace=False) == 4

    assert boundary_handling.set_boundary_where_flag_is_set(boundary_handling.boundary_objects[0], 16) == 16


def test_add_fix_steps():
    dh = SerialDataHandling(domain_size=(7, 7))
    src = dh.add_array('src')
    dh.fill("src", 0.0, ghost_layers=True)
    dh.fill("src", 1.0, ghost_layers=False)
    boundary_stencil = [(1, 0), (-1, 0), (0, 1), (0, -1)]

    boundary_handling = BoundaryHandling(dh, src.name, boundary_stencil,
                                         name="boundary_handling", target=pystencils.Target.CPU)

    neumann = Neumann()
    for d in ('N', 'S', 'W', 'E'):
        boundary_handling.set_boundary(neumann, slice_from_direction(d, dim=2))

    timeloop = TimeLoop(steps=1)
    boundary_handling.add_fixed_steps(timeloop)

    timeloop.run()
    assert np.sum(dh.cpu_arrays['src']) == 7 * 7 + 7 * 4


def test_boundary_data_setter():
    dh = SerialDataHandling(domain_size=(7, 7))
    src = dh.add_array('src')
    dh.fill("src", 0.0, ghost_layers=True)
    dh.fill("src", 1.0, ghost_layers=False)
    boundary_stencil = [(1, 0), (-1, 0), (0, 1), (0, -1)]

    boundary_handling = BoundaryHandling(dh, src.name, boundary_stencil,
                                         name="boundary_handling", target=Target.CPU)

    neumann = Neumann()
    for d in 'N':
        boundary_handling.set_boundary(neumann, slice_from_direction(d, dim=2))

    boundary_handling.prepare()

    for b in dh.iterate(ghost_layers=True):
        index_array_bd = b[boundary_handling._index_array_name]
        data_setter = index_array_bd.boundary_object_to_data_setter[boundary_handling.boundary_objects[0]]

        y_pos = data_setter.boundary_cell_positions(1)

        assert all(y_pos == 5.5)
        assert np.all(data_setter.link_offsets() == [0, -1])
        assert np.all(data_setter.link_positions(1) == 6.)


@pytest.mark.parametrize('with_indices', ('with_indices', False))
def test_dirichlet(with_indices):
    value = (1, 20, 3) if with_indices else 1

    dh = SerialDataHandling(domain_size=(7, 7))
    src = dh.add_array('src', values_per_cell=3 if with_indices else 1)
    dh.cpu_arrays.src[...] = np.random.rand(*src.shape)
    boundary_stencil = [(1, 0), (-1, 0), (0, 1), (0, -1)]
    boundary_handling = BoundaryHandling(dh, src.name, boundary_stencil)
    dirichlet = Dirichlet(value)
    assert dirichlet.name == 'Dirichlet'
    dirichlet.name = "wall"
    assert dirichlet.name == 'wall'

    for d in ('N', 'S', 'W', 'E'):
        boundary_handling.set_boundary(dirichlet, slice_from_direction(d, dim=2))
    boundary_handling()

    assert all([np.allclose(a, np.array(value)) for a in dh.cpu_arrays.src[1:-2, 0]])
    assert all([np.allclose(a, np.array(value)) for a in dh.cpu_arrays.src[1:-2, -1]])
    assert all([np.allclose(a, np.array(value)) for a in dh.cpu_arrays.src[0, 1:-2]])
    assert all([np.allclose(a, np.array(value)) for a in dh.cpu_arrays.src[-1, 1:-2]])



import numpy as np
import sympy as sp
from sympy.logic.boolalg import Boolean

from pystencils.typing.types import AbstractType, BasicType
from pystencils.typing.typed_sympy import TypedSymbol


class CastFunc(sp.Function):
    """
    CastFunc is used in order to introduce static casts. They are especially useful as a way to signal what type
    a certain node should have, if it is impossible to add a type to a node, e.g. a sp.Number.
    """
    is_Atom = True

    def __new__(cls, *args, **kwargs):
        if len(args) != 2:
            pass
        expr, dtype, *other_args = args

        # If we have two consecutive casts, throw the inner one away.
        # This optimisation is only available for simple casts. Thus the == is intended here!
        if expr.__class__ == CastFunc:
            expr = expr.args[0]
        if not isinstance(dtype, AbstractType):
            dtype = BasicType(dtype)
        # to work in conditions of sp.Piecewise cast_func has to be of type Boolean as well
        # however, a cast_function should only be a boolean if its argument is a boolean, otherwise this leads
        # to problems when for example comparing cast_func's for equality
        #
        # lhs = bitwise_and(a, cast_func(1, 'int'))
        # rhs = cast_func(0, 'int')
        # print( sp.Ne(lhs, rhs) ) # would give true if all cast_funcs are booleans
        # -> thus a separate class boolean_cast_func is introduced
        if isinstance(expr, Boolean) and (not isinstance(expr, TypedSymbol) or expr.dtype == BasicType('bool')):
            cls = BooleanCastFunc

        return sp.Function.__new__(cls, expr, dtype, *other_args, **kwargs)

    @property
    def canonical(self):
        if hasattr(self.args[0], 'canonical'):
            return self.args[0].canonical
        else:
            raise NotImplementedError()

    @property
    def is_commutative(self):
        return self.args[0].is_commutative

    @property
    def dtype(self):
        return self.args[1]

    @property
    def expr(self):
        return self.args[0]

    @property
    def is_integer(self):
        """
        Uses Numpy type hierarchy to determine :func:`sympy.Expr.is_integer` predicate

        For reference: Numpy type hierarchy https://docs.scipy.org/doc/numpy-1.13.0/reference/arrays.scalars.html
        """
        if hasattr(self.dtype, 'numpy_dtype'):
            return np.issubdtype(self.dtype.numpy_dtype, np.integer) or super().is_integer
        else:
            return super().is_integer

    @property
    def is_negative(self):
        """
        See :func:`.TypedSymbol.is_integer`
        """
        if hasattr(self.dtype, 'numpy_dtype'):
            if np.issubdtype(self.dtype.numpy_dtype, np.unsignedinteger):
                return False

        return super().is_negative

    @property
    def is_nonnegative(self):
        """
        See :func:`.TypedSymbol.is_integer`
        """
        if self.is_negative is False:
            return True
        else:
            return super().is_nonnegative

    @property
    def is_real(self):
        """
        See :func:`.TypedSymbol.is_integer`
        """
        if hasattr(self.dtype, 'numpy_dtype'):
            return np.issubdtype(self.dtype.numpy_dtype, np.integer) or np.issubdtype(self.dtype.numpy_dtype,
                                                                                      np.floating) or super().is_real
        else:
            return super().is_real


class BooleanCastFunc(CastFunc, Boolean):
    # TODO: documentation
    pass


class VectorMemoryAccess(CastFunc):
    """
    Special memory access for vectorized kernel.
    Arguments: read/write expression, type, aligned, non-temporal, mask (or none), stride
    """
    nargs = (6,)


class ReinterpretCastFunc(CastFunc):
    """
    Reinterpret cast is necessary for the StructType
    """
    pass


class PointerArithmeticFunc(sp.Function, Boolean):
    # TODO: documentation, or deprecate!
    @property
    def canonical(self):
        if hasattr(self.args[0], 'canonical'):
            return self.args[0].canonical
        else:
            raise NotImplementedError()



from pystencils.cpu.cpujit import make_python_function
from pystencils.cpu.kernelcreation import add_openmp, create_indexed_kernel, create_kernel

__all__ = ['create_kernel', 'create_indexed_kernel', 'add_openmp', 'make_python_function']



from typing import List

from pystencils.config import CreateKernelConfig
from pystencils.typing.leaf_typing import TypeAdder
from sympy.codegen import Assignment


def add_types(eqs: List[Assignment], config: CreateKernelConfig):
    """Traverses AST and replaces every :class:`sympy.Symbol` by a :class:`pystencils.typedsymbol.TypedSymbol`.

    Additionally returns sets of all fields which are read/written

    Args:
        eqs: list of equations
        config: CreateKernelConfig

    Returns:
        ``typed_equations`` list of equations where symbols have been replaced by typed symbols
    """

    check = TypeAdder(type_for_symbol=config.data_type,
                      default_number_float=config.default_number_float,
                      default_number_int=config.default_number_int)

    return check.visit(eqs)



import numpy as np

from pystencils import get_code_obj
from pystencils.astnodes import Block, KernelFunction, SympyAssignment
from pystencils.cpu import make_python_function
from pystencils.field import Field
from pystencils.enums import Target, Backend
from pystencils.transformations import (
    make_loop_over_domain, move_constants_before_loop, resolve_field_accesses)


def test_jacobi_fixed_field_size():
    size = (30, 20)

    src_field_c = np.random.rand(*size)
    src_field_py = np.copy(src_field_c)
    dst_field_c = np.zeros(size)
    dst_field_py = np.zeros(size)

    f = Field.create_from_numpy_array("f", src_field_c)
    d = Field.create_from_numpy_array("d", dst_field_c)

    jacobi = SympyAssignment(d[0, 0], (f[1, 0] + f[-1, 0] + f[0, 1] + f[0, -1]) / 4)
    body = Block([jacobi])
    loop_node, gl_info = make_loop_over_domain(body)
    ast_node = KernelFunction(loop_node, Target.CPU, Backend.C, make_python_function, ghost_layers=gl_info)
    resolve_field_accesses(ast_node)
    move_constants_before_loop(ast_node)

    for x in range(1, size[0] - 1):
        for y in range(1, size[1] - 1):
            dst_field_py[x, y] = 0.25 * (src_field_py[x - 1, y] + src_field_py[x + 1, y] +
                                         src_field_py[x, y - 1] + src_field_py[x, y + 1])

    kernel = ast_node.compile()
    kernel(f=src_field_c, d=dst_field_c)
    error = np.sum(np.abs(dst_field_py - dst_field_c))
    np.testing.assert_allclose(error, 0.0, atol=1e-13)

    code_display = get_code_obj(ast_node)
    assert 'for' in str(code_display)
    assert 'for' in code_display._repr_html_()


def test_jacobi_variable_field_size():
    size = (3, 3, 3)
    f = Field.create_generic("f", 3)
    d = Field.create_generic("d", 3)
    jacobi = SympyAssignment(d[0, 0, 0], (f[1, 0, 0] + f[-1, 0, 0] + f[0, 1, 0] + f[0, -1, 0]) / 4)
    body = Block([jacobi])
    loop_node, gl_info = make_loop_over_domain(body)
    ast_node = KernelFunction(loop_node, Target.CPU, Backend.C, make_python_function, ghost_layers=gl_info)
    resolve_field_accesses(ast_node)
    move_constants_before_loop(ast_node)

    src_field_c = np.random.rand(*size)
    src_field_py = np.copy(src_field_c)
    dst_field_c = np.zeros(size)
    dst_field_py = np.zeros(size)

    for x in range(1, size[0] - 1):
        for y in range(1, size[1] - 1):
            for z in range(1, size[2] - 1):
                dst_field_py[x, y, z] = 0.25 * (src_field_py[x - 1, y, z] + src_field_py[x + 1, y, z] +
                                                src_field_py[x, y - 1, z] + src_field_py[x, y + 1, z])

    kernel = ast_node.compile()
    kernel(f=src_field_c, d=dst_field_c)
    error = np.sum(np.abs(dst_field_py - dst_field_c))
    np.testing.assert_allclose(error, 0.0, atol=1e-13)



import os
from collections.abc import Hashable
from functools import partial, wraps
from itertools import chain

from functools import lru_cache as memorycache

from joblib import Memory
from appdirs import user_cache_dir

if 'PYSTENCILS_CACHE_DIR' in os.environ:
    cache_dir = os.environ['PYSTENCILS_CACHE_DIR']
else:
    cache_dir = user_cache_dir('pystencils')
disk_cache = Memory(cache_dir, verbose=False).cache
disk_cache_no_fallback = disk_cache


def _wrapper(wrapped_func, cached_func, *args, **kwargs):
    if all(isinstance(a, Hashable) for a in chain(args, kwargs.values())):
        return cached_func(*args, **kwargs)
    else:
        return wrapped_func(*args, **kwargs)


def memorycache_if_hashable(maxsize=128, typed=False):
    def wrapper(func):
        return partial(_wrapper, func, memorycache(maxsize, typed)(func))

    return wrapper


def sharedmethodcache(cache_id: str):
    """Decorator for memoization of instance methods, allowing multiple methods to use the same cache.

    This decorator caches results of instance methods per instantiated object of the surrounding class.
    It allows multiple methods to use the same cache, by passing them the same `cache_id` string.
    Cached values are stored in a dictionary, which is added as a member `self.<cache_id>` to the 
    `self` object instance. Make sure that this doesn't cause any naming conflicts with other members!
    Of course, for this to be useful, said methods must have the same signature (up to additional kwargs)
    and must return the same result when called with the same arguments."""
    def _decorator(user_method):
        @wraps(user_method)
        def _decorated_func(self, *args, **kwargs):
            objdict = self.__dict__
            cache = objdict.setdefault(cache_id, dict())

            key = args
            for item in kwargs.items():
                key += item

            if key not in cache:
                result = user_method(self, *args, **kwargs)
                cache[key] = result
                return result
            else:
                return cache[key]
        return _decorated_func
    return _decorator


# Disable memory cache:
# disk_cache = lambda o: o
# disk_cache_no_fallback = lambda o: o



import numpy as np
from itertools import product

from pystencils import CreateKernelConfig, create_kernel
import pystencils.gpucuda
from pystencils import Assignment, Field
from pystencils.enums import Target
from pystencils.slicing import get_periodic_boundary_src_dst_slices, normalize_slice


def create_copy_kernel(domain_size, from_slice, to_slice, index_dimensions=0, index_dim_shape=1, dtype=np.float64):
    """Copies a rectangular part of an array to another non-overlapping part"""

    f = Field.create_generic("pdfs", len(domain_size), index_dimensions=index_dimensions, dtype=dtype)
    normalized_from_slice = normalize_slice(from_slice, f.spatial_shape)
    normalized_to_slice = normalize_slice(to_slice, f.spatial_shape)

    offset = [s1.start - s2.start for s1, s2 in zip(normalized_from_slice, normalized_to_slice)]
    assert offset == [s1.stop - s2.stop for s1, s2 in zip(normalized_from_slice, normalized_to_slice)], \
        "Slices have to have same size"

    update_eqs = []
    if index_dimensions < 2:
        index_dim_shape = [index_dim_shape]
    for i in product(*[range(d) for d in index_dim_shape]):
        eq = Assignment(f(*i), f[tuple(offset)](*i))
        update_eqs.append(eq)

    config = CreateKernelConfig(target=Target.GPU, iteration_slice=to_slice, skip_independence_check=True)

    ast = create_kernel(update_eqs, config=config)
    return ast


def get_periodic_boundary_functor(stencil, domain_size, index_dimensions=0, index_dim_shape=1, ghost_layers=1,
                                  thickness=None, dtype=np.float64, target=Target.GPU):
    assert target in {Target.GPU}
    src_dst_slice_tuples = get_periodic_boundary_src_dst_slices(stencil, ghost_layers, thickness)
    kernels = []

    for src_slice, dst_slice in src_dst_slice_tuples:
        ast = create_copy_kernel(domain_size, src_slice, dst_slice, index_dimensions, index_dim_shape, dtype)
        kernels.append(pystencils.gpucuda.make_python_function(ast))

    def functor(pdfs, **_):
        for kernel in kernels:
            kernel(pdfs=pdfs)

    return functor



from subprocess import CalledProcessError

import pytest

import pystencils
import pystencils.cpu.cpujit
from pystencils.backends.cbackend import CBackend
from pystencils.backends.cuda_backend import CudaBackend
from pystencils.enums import Target


class ScreamingBackend(CBackend):

    def _print(self, node):
        normal_code = super()._print(node)
        return normal_code.upper()


class ScreamingGpuBackend(CudaBackend):

    def _print(self, node):
        normal_code = super()._print(node)
        return normal_code.upper()


def test_custom_backends_cpu():
    z, y, x = pystencils.fields("z, y, x: [2d]")

    normal_assignments = pystencils.AssignmentCollection([pystencils.Assignment(
        z[0, 0], x[0, 0] * x[0, 0] * y[0, 0])], [])

    ast = pystencils.create_kernel(normal_assignments, target=Target.CPU)
    pystencils.show_code(ast, ScreamingBackend())
    with pytest.raises(CalledProcessError):
        pystencils.cpu.cpujit.make_python_function(ast, custom_backend=ScreamingBackend())


def test_custom_backends_gpu():
    pytest.importorskip('pycuda')
    import pycuda.driver
    import pystencils.gpucuda.cudajit

    z, x, y = pystencils.fields("z, y, x: [2d]")

    normal_assignments = pystencils.AssignmentCollection([pystencils.Assignment(
        z[0, 0], x[0, 0] * x[0, 0] * y[0, 0])], [])

    ast = pystencils.create_kernel(normal_assignments, target=Target.GPU)
    pystencils.show_code(ast, ScreamingGpuBackend())
    with pytest.raises(pycuda.driver.CompileError):
        pystencils.gpucuda.cudajit.make_python_function(ast, custom_backend=ScreamingGpuBackend())



import numpy as np

import pystencils as ps


def test_blocking_staggered():
    f = ps.fields("f: double[3D]")
    stag = ps.fields("stag(3): double[3D]", field_type=ps.FieldType.STAGGERED)
    terms = [
       f[0, 0, 0] - f[-1, 0, 0],
       f[0, 0, 0] - f[0, -1, 0],
       f[0, 0, 0] - f[0, 0, -1],
    ]
    assignments = [ps.Assignment(stag.staggered_access(d), terms[i]) for i, d in enumerate(stag.staggered_stencil)]
    reference_kernel = ps.create_staggered_kernel(assignments)
    print(ps.show_code(reference_kernel))
    reference_kernel = reference_kernel.compile()
    kernel = ps.create_staggered_kernel(assignments, cpu_blocking=(3, 16, 8)).compile()
    print(ps.show_code(kernel.ast))

    f_arr = np.random.rand(80, 33, 19)
    stag_arr = np.zeros((80, 33, 19, 3))
    stag_ref = np.zeros((80, 33, 19, 3))
    kernel(f=f_arr, stag=stag_arr)
    reference_kernel(f=f_arr, stag=stag_ref)
    np.testing.assert_almost_equal(stag_arr, stag_ref)



import numpy as np
import sympy as sp
from sympy.codegen.ast import Assignment
from sympy.printing.latex import LatexPrinter

__all__ = ['Assignment', 'assignment_from_stencil']


def print_assignment_latex(printer, expr):
    """sympy cannot print Assignments as Latex. Thus, this function is added to the sympy Latex printer"""
    printed_lhs = printer.doprint(expr.lhs)
    printed_rhs = printer.doprint(expr.rhs)
    return fr"{printed_lhs} \leftarrow {printed_rhs}"


def assignment_str(assignment):
    return fr"{assignment.lhs} ← {assignment.rhs}"


_old_new = sp.codegen.ast.Assignment.__new__


# TODO Typing Part2 add default type, defult_float_type, default_int_type and use sane defaults
def _Assignment__new__(cls, lhs, rhs, *args, **kwargs):
    if isinstance(lhs, (list, tuple, sp.Matrix)) and isinstance(rhs, (list, tuple, sp.Matrix)):
        assert len(lhs) == len(rhs), f'{lhs} and {rhs} must have same length when performing vector assignment!'
        return tuple(_old_new(cls, a, b, *args, **kwargs) for a, b in zip(lhs, rhs))
    return _old_new(cls, lhs, rhs, *args, **kwargs)


Assignment.__str__ = assignment_str
Assignment.__new__ = _Assignment__new__
LatexPrinter._print_Assignment = print_assignment_latex

sp.MutableDenseMatrix.__hash__ = lambda self: hash(tuple(self))


def assignment_from_stencil(stencil_array, input_field, output_field,
                            normalization_factor=None, order='visual') -> Assignment:
    """Creates an assignment

    Args:
        stencil_array: nested list of numpy array defining the stencil weights
        input_field: field or field access, defining where the stencil should be applied to
        output_field: field or field access where the result is written to
        normalization_factor: optional normalization factor for the stencil
        order: defines how the stencil_array is interpreted. Possible values are 'visual' and 'numpy'.
               For details see examples

    Returns:
        Assignment that can be used to create a kernel

    Examples:
        >>> import pystencils as ps
        >>> f, g = ps.fields("f, g: [2D]")
        >>> stencil = [[0, 2, 0],
        ...            [3, 4, 5],
        ...            [0, 6, 0]]

        By default 'visual ordering is used - i.e. the stencil is applied as the nested lists are written down
        >>> expected_output = Assignment(g[0, 0], 3*f[-1, 0] + 6*f[0, -1] + 4*f[0, 0] + 2*f[0, 1] + 5*f[1, 0])
        >>> assignment_from_stencil(stencil, f, g, order='visual') == expected_output
        True

        'numpy' ordering uses the first coordinate of the stencil array for x offset, second for y offset etc.
        >>> expected_output = Assignment(g[0, 0], 2*f[-1, 0] + 3*f[0, -1] + 4*f[0, 0] + 5*f[0, 1] + 6*f[1, 0])
        >>> assignment_from_stencil(stencil, f, g, order='numpy') == expected_output
        True

        You can also pass field accesses to apply the stencil at an already shifted position:
        >>> expected_output = Assignment(g[2, 0], 3*f[0, 0] + 6*f[1, -1] + 4*f[1, 0] + 2*f[1, 1] + 5*f[2, 0])
        >>> assignment_from_stencil(stencil, f[1, 0], g[2, 0]) == expected_output
        True
    """
    from pystencils.field import Field

    stencil_array = np.array(stencil_array)
    if order == 'visual':
        stencil_array = np.swapaxes(stencil_array, 0, 1)
        stencil_array = np.flip(stencil_array, axis=1)
    elif order == 'numpy':
        pass
    else:
        raise ValueError("'order' has to be either 'visual' or 'numpy'")

    if isinstance(input_field, Field):
        input_field = input_field.center
    if isinstance(output_field, Field):
        output_field = output_field.center

    rhs = 0
    offset = tuple(s // 2 for s in stencil_array.shape)

    for index, factor in np.ndenumerate(stencil_array):
        shift = tuple(i - o for i, o in zip(index, offset))
        rhs += factor * input_field.get_shifted(*shift)

    if normalization_factor:
        rhs *= normalization_factor

    return Assignment(output_field, rhs)



import abc
from functools import partial

import sympy as sp
from sympy.core.cache import cacheit

from pystencils.astnodes import Block, Conditional
from pystencils.typing import TypedSymbol, create_type
from pystencils.integer_functions import div_ceil, div_floor
from pystencils.slicing import normalize_slice
from pystencils.sympyextensions import is_integer_sequence, prod


class ThreadIndexingSymbol(TypedSymbol):
    def __new__(cls, *args, **kwds):
        obj = ThreadIndexingSymbol.__xnew_cached_(cls, *args, **kwds)
        return obj

    def __new_stage2__(cls, name, dtype, *args, **kwargs):
        obj = super(ThreadIndexingSymbol, cls).__xnew__(cls, name, dtype, *args, **kwargs)
        return obj

    __xnew__ = staticmethod(__new_stage2__)
    __xnew_cached_ = staticmethod(cacheit(__new_stage2__))


BLOCK_IDX = [ThreadIndexingSymbol("blockIdx." + coord, create_type("int32")) for coord in ('x', 'y', 'z')]
THREAD_IDX = [ThreadIndexingSymbol("threadIdx." + coord, create_type("int32")) for coord in ('x', 'y', 'z')]
BLOCK_DIM = [ThreadIndexingSymbol("blockDim." + coord, create_type("int32")) for coord in ('x', 'y', 'z')]
GRID_DIM = [ThreadIndexingSymbol("gridDim." + coord, create_type("int32")) for coord in ('x', 'y', 'z')]


class AbstractIndexing(abc.ABC):
    """
    Abstract base class for all Indexing classes. An Indexing class defines how a multidimensional
    field is mapped to CUDA's block and grid system. It calculates indices based on CUDA's thread and block indices
    and computes the number of blocks and threads a kernel is started with. The Indexing class is created with
    a pystencils field, a slice to iterate over, and further optional parameters that must have default values.
    """

    @property
    @abc.abstractmethod
    def coordinates(self):
        """Returns a sequence of coordinate expressions for (x,y,z) depending on symbolic CUDA block and thread indices.
        These symbolic indices can be obtained with the method `index_variables` """

    @property
    def index_variables(self):
        """Sympy symbols for CUDA's block and thread indices, and block and grid dimensions. """
        return BLOCK_IDX + THREAD_IDX + BLOCK_DIM + GRID_DIM

    @abc.abstractmethod
    def call_parameters(self, arr_shape):
        """Determine grid and block size for kernel call.

        Args:
            arr_shape: the numeric (not symbolic) shape of the array
        Returns:
            dict with keys 'blocks' and 'threads' with tuple values for number of (x,y,z) threads and blocks
            the kernel should be started with
        """

    @abc.abstractmethod
    def guard(self, kernel_content, arr_shape):
        """In some indexing schemes not all threads of a block execute the kernel content.

        This function can return a Conditional ast node, defining this execution guard.

        Args:
            kernel_content: the actual kernel contents which can e.g. be put into the Conditional node as true block
            arr_shape: the numeric or symbolic shape of the field

        Returns:
            ast node, which is put inside the kernel function
        """

    @abc.abstractmethod
    def max_threads_per_block(self):
        """Return maximal number of threads per block for launch bounds. If this cannot be determined without
        knowing the array shape return None for unknown """

    @abc.abstractmethod
    def symbolic_parameters(self):
        """Set of symbols required in call_parameters code"""


# -------------------------------------------- Implementations ---------------------------------------------------------


class BlockIndexing(AbstractIndexing):
    """Generic indexing scheme that maps sub-blocks of an array to CUDA blocks.

    Args:
        field: pystencils field (common to all Indexing classes)
        iteration_slice: slice that defines rectangular subarea which is iterated over
        permute_block_size_dependent_on_layout: if True the block_size is permuted such that the fastest coordinate
                                                gets the largest amount of threads
        compile_time_block_size: compile in concrete block size, otherwise the cuda variable 'blockDim' is used
    """

    def __init__(self, field, iteration_slice,
                 block_size=(16, 16, 1), permute_block_size_dependent_on_layout=True, compile_time_block_size=False,
                 maximum_block_size=(1024, 1024, 64)):
        if field.spatial_dimensions > 3:
            raise NotImplementedError("This indexing scheme supports at most 3 spatial dimensions")

        if permute_block_size_dependent_on_layout:
            block_size = self.permute_block_size_according_to_layout(block_size, field.layout)

        self._block_size = block_size
        if maximum_block_size == 'auto':
            # Get device limits
            import pycuda.driver as cuda
            # noinspection PyUnresolvedReferences
            import pycuda.autoinit  # NOQA
            da = cuda.device_attribute
            device = cuda.Context.get_device()
            maximum_block_size = tuple(device.get_attribute(a)
                                       for a in (da.MAX_BLOCK_DIM_X, da.MAX_BLOCK_DIM_Y, da.MAX_BLOCK_DIM_Z))

        self._maximum_block_size = maximum_block_size
        self._iterationSlice = normalize_slice(iteration_slice, field.spatial_shape)
        self._dim = field.spatial_dimensions
        self._symbolic_shape = [e if isinstance(e, sp.Basic) else None for e in field.spatial_shape]
        self._compile_time_block_size = compile_time_block_size

    @property
    def coordinates(self):
        offsets = _get_start_from_slice(self._iterationSlice)
        block_size = self._block_size if self._compile_time_block_size else BLOCK_DIM
        coordinates = [block_index * bs + thread_idx + off
                       for block_index, bs, thread_idx, off in zip(BLOCK_IDX, block_size, THREAD_IDX, offsets)]

        return coordinates[:self._dim]

    def call_parameters(self, arr_shape):
        substitution_dict = {sym: value for sym, value in zip(self._symbolic_shape, arr_shape) if sym is not None}

        widths = [end - start for start, end in zip(_get_start_from_slice(self._iterationSlice),
                                                    _get_end_from_slice(self._iterationSlice, arr_shape))]
        widths = sp.Matrix(widths).subs(substitution_dict)
        extend_bs = (1,) * (3 - len(self._block_size))
        block_size = self._block_size + extend_bs
        if not self._compile_time_block_size:
            assert len(block_size) == 3
            adapted_block_size = []
            for i in range(len(widths)):
                factor = div_floor(prod(block_size[:i]), prod(adapted_block_size))
                adapted_block_size.append(sp.Min(block_size[i] * factor, widths[i]))
            extend_adapted_bs = (1,) * (3 - len(adapted_block_size))
            block_size = tuple(adapted_block_size) + extend_adapted_bs

        block_size = tuple(sp.Min(bs, max_bs) for bs, max_bs in zip(block_size, self._maximum_block_size))
        grid = tuple(div_ceil(length, block_size) for length, block_size in zip(widths, block_size))
        extend_gr = (1,) * (3 - len(grid))

        return {'block': block_size,
                'grid': grid + extend_gr}

    def guard(self, kernel_content, arr_shape):
        arr_shape = arr_shape[:self._dim]
        conditions = [c < end
                      for c, end in zip(self.coordinates, _get_end_from_slice(self._iterationSlice, arr_shape))]
        condition = conditions[0]
        for c in conditions[1:]:
            condition = sp.And(condition, c)
        return Block([Conditional(condition, kernel_content)])

    @staticmethod
    def limit_block_size_by_register_restriction(block_size, required_registers_per_thread, device=None):
        """Shrinks the block_size if there are too many registers used per multiprocessor.
        This is not done automatically, since the required_registers_per_thread are not known before compilation.
        They can be obtained by ``func.num_regs`` from a pycuda function.
        :returns smaller block_size if too many registers are used.
        """
        import pycuda.driver as cuda
        # noinspection PyUnresolvedReferences
        import pycuda.autoinit  # NOQA

        da = cuda.device_attribute
        if device is None:
            device = cuda.Context.get_device()
        available_registers_per_mp = device.get_attribute(da.MAX_REGISTERS_PER_MULTIPROCESSOR)

        block = block_size

        while True:
            num_threads = 1
            for t in block:
                num_threads *= t
            required_registers_per_mt = num_threads * required_registers_per_thread
            if required_registers_per_mt <= available_registers_per_mp:
                return block
            else:
                largest_grid_entry_idx = max(range(len(block)), key=lambda e: block[e])
                assert block[largest_grid_entry_idx] >= 2
                block[largest_grid_entry_idx] //= 2

    @staticmethod
    def permute_block_size_according_to_layout(block_size, layout):
        """Returns modified block_size such that the fastest coordinate gets the biggest block dimension"""
        if not is_integer_sequence(block_size):
            return block_size
        sorted_block_size = list(sorted(block_size, reverse=True))
        while len(sorted_block_size) > len(layout):
            sorted_block_size[0] *= sorted_block_size[-1]
            sorted_block_size = sorted_block_size[:-1]

        result = list(block_size)
        for l, bs in zip(reversed(layout), sorted_block_size):
            result[l] = bs
        return tuple(result[:len(layout)])

    def max_threads_per_block(self):
        if is_integer_sequence(self._block_size):
            return prod(self._block_size)
        else:
            return None

    def symbolic_parameters(self):
        return set(b for b in self._block_size if isinstance(b, sp.Symbol))


class LineIndexing(AbstractIndexing):
    """
    Indexing scheme that assigns the innermost 'line' i.e. the elements which are adjacent in memory to a 1D CUDA block.
    The fastest coordinate is indexed with thread_idx.x, the remaining coordinates are mapped to block_idx.{x,y,z}
    This indexing scheme supports up to 4 spatial dimensions, where the innermost dimensions is not larger than the
    maximum amount of threads allowed in a CUDA block (which depends on device).
    """

    def __init__(self, field, iteration_slice):
        available_indices = [THREAD_IDX[0]] + BLOCK_IDX
        if field.spatial_dimensions > 4:
            raise NotImplementedError("This indexing scheme supports at most 4 spatial dimensions")

        coordinates = available_indices[:field.spatial_dimensions]

        fastest_coordinate = field.layout[-1]
        coordinates[0], coordinates[fastest_coordinate] = coordinates[fastest_coordinate], coordinates[0]

        self._coordinates = coordinates
        self._iterationSlice = normalize_slice(iteration_slice, field.spatial_shape)
        self._symbolicShape = [e if isinstance(e, sp.Basic) else None for e in field.spatial_shape]

    @property
    def coordinates(self):
        return [i + offset for i, offset in zip(self._coordinates, _get_start_from_slice(self._iterationSlice))]

    def call_parameters(self, arr_shape):
        substitution_dict = {sym: value for sym, value in zip(self._symbolicShape, arr_shape) if sym is not None}

        widths = [end - start for start, end in zip(_get_start_from_slice(self._iterationSlice),
                                                    _get_end_from_slice(self._iterationSlice, arr_shape))]
        widths = sp.Matrix(widths).subs(substitution_dict)

        def get_shape_of_cuda_idx(cuda_idx):
            if cuda_idx not in self._coordinates:
                return 1
            else:
                idx = self._coordinates.index(cuda_idx)
                return widths[idx]

        return {'block': tuple([get_shape_of_cuda_idx(idx) for idx in THREAD_IDX]),
                'grid': tuple([get_shape_of_cuda_idx(idx) for idx in BLOCK_IDX])}

    def guard(self, kernel_content, arr_shape):
        return kernel_content

    def max_threads_per_block(self):
        return None

    def symbolic_parameters(self):
        return set()


# -------------------------------------- Helper functions --------------------------------------------------------------

def _get_start_from_slice(iteration_slice):
    res = []
    for slice_component in iteration_slice:
        if type(slice_component) is slice:
            res.append(slice_component.start if slice_component.start is not None else 0)
        else:
            assert isinstance(slice_component, int)
            res.append(slice_component)
    return res


def _get_end_from_slice(iteration_slice, arr_shape):
    iter_slice = normalize_slice(iteration_slice, arr_shape)
    res = []
    for slice_component in iter_slice:
        if type(slice_component) is slice:
            res.append(slice_component.stop)
        else:
            assert isinstance(slice_component, int)
            res.append(slice_component + 1)
    return res


def indexing_creator_from_params(gpu_indexing, gpu_indexing_params):
    if isinstance(gpu_indexing, str):
        if gpu_indexing == 'block':
            indexing_creator = BlockIndexing
        elif gpu_indexing == 'line':
            indexing_creator = LineIndexing
        else:
            raise ValueError(f"Unknown GPU indexing {gpu_indexing}. Valid values are 'block' and 'line'")
        if gpu_indexing_params:
            indexing_creator = partial(indexing_creator, **gpu_indexing_params)
        return indexing_creator
    else:
        return gpu_indexing



#
# Copyright © 2020 Stephan Seitz <stephan.seitz@fau.de>
#
# Distributed under terms of the GPLv3 license.

"""

"""

import pytest
from pystencils.session import *

from sympy import poly


def test_field_access_poly():
    dh = ps.create_data_handling((20, 20))
    ρ = dh.add_array('rho')
    rho = ρ.center
    a = poly(rho+0.5, rho)
    print(a)


def test_field_access_piecewise():
    try:
        a = sp.Piecewise((0, 1 < sp.Max(-0.5, sp.Symbol("test") + 0.5)), (1, True))
        a.simplify()
    except Exception as e:
        pytest.skip(f"Bug in SymPy 1.10: {e}")
    else:
        dh = ps.create_data_handling((20, 20))
        ρ = dh.add_array('rho')
        pw = sp.Piecewise((0, 1 < sp.Max(-0.5, ρ.center+0.5)), (1, True))
        a = sp.simplify(pw)
        print(a)



import sympy as sp

from pystencils import AssignmentCollection, Assignment
from pystencils.node_collection import NodeCollection
from pystencils.astnodes import SympyAssignment


def test_node_collection_from_assignment_collection():
    x = sp.symbols('x')
    assignment_collection = AssignmentCollection([Assignment(x, 2)])
    node_collection = NodeCollection.from_assignment_collection(assignment_collection)

    assert node_collection.all_assignments[0] == SympyAssignment(x, 2)



import pytest
import sympy as sp
import numpy as np
import pystencils as ps
from pystencils.fast_approximation import fast_division


@pytest.mark.parametrize('dtype', ["float64", "float32"])
@pytest.mark.parametrize('func', [sp.Pow, sp.atan2])
@pytest.mark.parametrize('target', [ps.Target.CPU, ps.Target.GPU])
def test_two_arguments(dtype, func, target):
    if target == ps.Target.GPU:
        pytest.importorskip("pycuda")
    dh = ps.create_data_handling(domain_size=(10, 10), periodicity=True, default_target=target)

    x = dh.add_array('x', values_per_cell=1, dtype=dtype)
    dh.fill("x", 0.0, ghost_layers=True)
    y = dh.add_array('y', values_per_cell=1, dtype=dtype)
    dh.fill("y", 1.0, ghost_layers=True)
    z = dh.add_array('z', values_per_cell=1, dtype=dtype)
    dh.fill("z", 2.0, ghost_layers=True)

    config = ps.CreateKernelConfig(target=target)

    # test sp.Max with one argument
    up = ps.Assignment(x.center, func(y.center, z.center))
    ast = ps.create_kernel(up, config=config)
    code = ps.get_code_str(ast)
    if dtype == 'float32':
        assert func.__name__.lower() in code
    kernel = ast.compile()

    dh.all_to_gpu()
    dh.run_kernel(kernel)
    dh.all_to_cpu()

    np.testing.assert_allclose(dh.gather_array("x")[0, 0], float(func(1.0, 2.0).evalf()),
                               13 if dtype == 'float64' else 5)


@pytest.mark.parametrize('dtype', ["float64", "float32"])
@pytest.mark.parametrize('func', [sp.sin, sp.cos, sp.sinh, sp.cosh, sp.atan])
@pytest.mark.parametrize('target', [ps.Target.CPU, ps.Target.GPU])
def test_single_arguments(dtype, func, target):
    if target == ps.Target.GPU:
        pytest.importorskip("pycuda")
    dh = ps.create_data_handling(domain_size=(10, 10), periodicity=True, default_target=target)

    x = dh.add_array('x', values_per_cell=1, dtype=dtype)
    dh.fill("x", 0.0, ghost_layers=True)
    y = dh.add_array('y', values_per_cell=1, dtype=dtype)
    dh.fill("y", 1.0, ghost_layers=True)

    config = ps.CreateKernelConfig(target=target)

    # test sp.Max with one argument
    up = ps.Assignment(x.center, func(y.center))
    ast = ps.create_kernel(up, config=config)
    code = ps.get_code_str(ast)
    if dtype == 'float32':
        assert func.__name__.lower() in code
    kernel = ast.compile()

    dh.all_to_gpu()
    dh.run_kernel(kernel)
    dh.all_to_cpu()

    np.testing.assert_allclose(dh.gather_array("x")[0, 0], float(func(1.0).evalf()),
                               rtol=10**-3 if dtype == 'float32' else 10**-5)


@pytest.mark.parametrize('a', [sp.Symbol('a'), ps.fields('a:  float64[2d]').center])
def test_avoid_pow(a):
    x = ps.fields('x:  float64[2d]')

    up = ps.Assignment(x.center_vector[0], 2 * a ** 2 / 3)
    ast = ps.create_kernel(up)

    code = ps.get_code_str(ast)

    assert "pow" not in code


def test_avoid_pow_fast_div():
    x = ps.fields('x:  float64[2d]')
    a = ps.fields('a:  float64[2d]').center

    up = ps.Assignment(x.center_vector[0], fast_division(1, (a**2)))
    ast = ps.create_kernel(up, config=ps.CreateKernelConfig(target=ps.Target.GPU))
    # ps.show_code(ast)

    code = ps.get_code_str(ast)

    assert "pow" not in code


def test_avoid_pow_move_constants():
    # At the end of the kernel creation the function move_constants_before_loop will be called
    # This function additionally contains substitutions for symbols with the same value
    # Thus it simplifies the equations again
    x = ps.fields('x:  float64[2d]')
    a, b, c = sp.symbols("a, b, c")

    up = [ps.Assignment(a, 0.0),
          ps.Assignment(b, 0.0),
          ps.Assignment(c, 0.0),
          ps.Assignment(x.center_vector[0], a**2/18 - a*b/6 - a/18 + b**2/18 + b/18 - c**2/36)]
    ast = ps.create_kernel(up)

    code = ps.get_code_str(ast)
    ps.show_code(ast)

    assert "pow" not in code



import os
from tempfile import TemporaryDirectory
import shutil

import pytest

import numpy as np

import pystencils.plot as plt


def example_scalar_field(t=0):
    x, y = np.meshgrid(np.linspace(0, 2 * np.pi, 100), np.linspace(0, 2 * np.pi, 100))
    z = np.cos(x + 0.1 * t) * np.sin(y + 0.1 * t) + 0.1 * x * y
    return z


def example_vector_field(t=0, shape=(40, 40)):
    result = np.empty(shape + (2,))
    x, y = np.meshgrid(np.linspace(0, 2 * np.pi, shape[0]), np.linspace(0, 2 * np.pi, shape[1]))
    result[..., 0] = np.cos(x + 0.1 * t) * np.sin(y + 0.1 * t) + 0.01 * x * y
    result[..., 1] = np.cos(0.001 * y)
    return result


@pytest.mark.skipif(shutil.which('ffmpeg') is None, reason="ffmpeg not available")
def test_animation():
    t = 0

    def run_scalar():
        nonlocal t
        t += 1
        return example_scalar_field(t)

    def run_vec():
        nonlocal t
        t += 1
        return example_vector_field(t)

    plt.clf()
    plt.cla()

    with TemporaryDirectory() as tmp_dir:
        ani = plt.vector_field_magnitude_animation(run_vec, interval=1, frames=2)
        ani.save(os.path.join(tmp_dir, "animation1.avi"))

        ani = plt.vector_field_animation(run_vec, interval=1, frames=2, rescale=True)
        ani.save(os.path.join(tmp_dir, "animation2.avi"))

        ani = plt.vector_field_animation(run_vec, interval=1, frames=2, rescale=False)
        ani.save(os.path.join(tmp_dir, "animation3.avi"))

        ani = plt.scalar_field_animation(run_scalar, interval=1, frames=2, rescale=True)
        ani.save(os.path.join(tmp_dir, "animation4.avi"))

        ani = plt.scalar_field_animation(run_scalar, interval=1, frames=2, rescale=False)
        ani.save(os.path.join(tmp_dir, "animation5.avi"))

        ani = plt.surface_plot_animation(run_scalar, frames=2)
        ani.save(os.path.join(tmp_dir, "animation6.avi"))



import pytest

import pystencils.config
import sympy

import pystencils as ps
from pystencils.typing import CastFunc, create_type


@pytest.mark.parametrize('target', (ps.Target.CPU, ps.Target.GPU))
def test_abs(target):
    x, y, z = ps.fields('x, y, z:  float64[2d]')

    default_int_type = create_type('int64')

    assignments = ps.AssignmentCollection({x[0, 0]: sympy.Abs(CastFunc(y[0, 0], default_int_type))})

    config = pystencils.config.CreateKernelConfig(target=target)
    ast = ps.create_kernel(assignments, config=config)
    code = ps.get_code_str(ast)
    print(code)
    assert 'fabs(' not in code



from typing import Union

import numpy as np
import sympy as sp
from sympy.core.cache import cacheit

from pystencils.typing.types import BasicType, create_type, PointerType


def assumptions_from_dtype(dtype: Union[BasicType, np.dtype]):
    """Derives SymPy assumptions from :class:`BasicType` or a Numpy dtype

    Args:
        dtype (BasicType, np.dtype): a Numpy data type
    Returns:
        A dict of SymPy assumptions
    """
    if hasattr(dtype, 'numpy_dtype'):
        dtype = dtype.numpy_dtype

    assumptions = dict()

    try:
        if np.issubdtype(dtype, np.integer):
            assumptions.update({'integer': True})

        if np.issubdtype(dtype, np.unsignedinteger):
            assumptions.update({'negative': False})

        if np.issubdtype(dtype, np.integer) or \
                np.issubdtype(dtype, np.floating):
            assumptions.update({'real': True})
    except Exception:  # TODO this is dirty
        pass

    return assumptions


class TypedSymbol(sp.Symbol):
    def __new__(cls, *args, **kwds):
        obj = TypedSymbol.__xnew_cached_(cls, *args, **kwds)
        return obj

    def __new_stage2__(cls, name, dtype, **kwargs):  # TODO does not match signature of sp.Symbol???
        # TODO: also Symbol should be allowed  ---> see sympy Variable
        assumptions = assumptions_from_dtype(dtype)
        assumptions.update(kwargs)
        obj = super(TypedSymbol, cls).__xnew__(cls, name, **assumptions)
        try:
            obj.numpy_dtype = create_type(dtype)
        except (TypeError, ValueError):
            # on error keep the string
            obj.numpy_dtype = dtype
        return obj

    __xnew__ = staticmethod(__new_stage2__)
    __xnew_cached_ = staticmethod(cacheit(__new_stage2__))

    @property
    def dtype(self):
        return self.numpy_dtype

    def _hashable_content(self):
        return super()._hashable_content(), hash(self.numpy_dtype)

    def __getnewargs__(self):
        return self.name, self.dtype

    def __getnewargs_ex__(self):
        return (self.name, self.dtype), self.assumptions0

    @property
    def canonical(self):
        return self

    @property
    def reversed(self):
        return self

    @property
    def headers(self):
        headers = []
        try:
            if np.issubdtype(self.dtype.numpy_dtype, np.complexfloating):
                headers.append('"cuda_complex.hpp"')
        except Exception:
            pass
        try:
            if np.issubdtype(self.dtype.base_type.numpy_dtype, np.complexfloating):
                headers.append('"cuda_complex.hpp"')
        except Exception:
            pass

        return headers


SHAPE_DTYPE = BasicType('int64', const=True)
STRIDE_DTYPE = BasicType('int64', const=True)


class FieldStrideSymbol(TypedSymbol):
    """Sympy symbol representing the stride value of a field in a specific coordinate."""
    def __new__(cls, *args, **kwds):
        obj = FieldStrideSymbol.__xnew_cached_(cls, *args, **kwds)
        return obj

    def __new_stage2__(cls, field_name, coordinate):
        name = f"_stride_{field_name}_{coordinate}"
        obj = super(FieldStrideSymbol, cls).__xnew__(cls, name, STRIDE_DTYPE, positive=True)
        obj.field_name = field_name
        obj.coordinate = coordinate
        return obj

    def __getnewargs__(self):
        return self.field_name, self.coordinate

    def __getnewargs_ex__(self):
        return (self.field_name, self.coordinate), {}

    __xnew__ = staticmethod(__new_stage2__)
    __xnew_cached_ = staticmethod(cacheit(__new_stage2__))

    def _hashable_content(self):
        return super()._hashable_content(), self.coordinate, self.field_name


class FieldShapeSymbol(TypedSymbol):
    """Sympy symbol representing the shape value of a sequence of fields. In a kernel iterating over multiple fields
    there is only one set of `FieldShapeSymbol`s since all the fields have to be of equal size."""
    def __new__(cls, *args, **kwds):
        obj = FieldShapeSymbol.__xnew_cached_(cls, *args, **kwds)
        return obj

    def __new_stage2__(cls, field_names, coordinate):
        names = "_".join([field_name for field_name in field_names])
        name = f"_size_{names}_{coordinate}"
        obj = super(FieldShapeSymbol, cls).__xnew__(cls, name, SHAPE_DTYPE, positive=True)
        obj.field_names = tuple(field_names)
        obj.coordinate = coordinate
        return obj

    def __getnewargs__(self):
        return self.field_names, self.coordinate

    def __getnewargs_ex__(self):
        return (self.field_names, self.coordinate), {}

    __xnew__ = staticmethod(__new_stage2__)
    __xnew_cached_ = staticmethod(cacheit(__new_stage2__))

    def _hashable_content(self):
        return super()._hashable_content(), self.coordinate, self.field_names


class FieldPointerSymbol(TypedSymbol):
    """Sympy symbol representing the pointer to the beginning of the field data."""
    def __new__(cls, *args, **kwds):
        obj = FieldPointerSymbol.__xnew_cached_(cls, *args, **kwds)
        return obj

    def __new_stage2__(cls, field_name, field_dtype, const):
        from pystencils.typing.utilities import get_base_type

        name = f"_data_{field_name}"
        dtype = PointerType(get_base_type(field_dtype), const=const, restrict=True)
        obj = super(FieldPointerSymbol, cls).__xnew__(cls, name, dtype)
        obj.field_name = field_name
        return obj

    def __getnewargs__(self):
        return self.field_name, self.dtype, self.dtype.const

    def __getnewargs_ex__(self):
        return (self.field_name, self.dtype, self.dtype.const), {}

    def _hashable_content(self):
        return super()._hashable_content(), self.field_name

    __xnew__ = staticmethod(__new_stage2__)
    __xnew_cached_ = staticmethod(cacheit(__new_stage2__))



from abc import abstractmethod
from typing import Union

import numpy as np
import sympy as sp


def is_supported_type(dtype: np.dtype):
    scalar = dtype.type
    c = np.issctype(dtype)
    subclass = issubclass(scalar, np.floating) or issubclass(scalar, np.integer) or issubclass(scalar, np.bool_)
    additional_checks = dtype.fields is None and dtype.hasobject is False and dtype.subdtype is None
    return c and subclass and additional_checks


def numpy_name_to_c(name: str) -> str:
    """
    Converts a np.dtype.name into a C type
    Args:
        name: np.dtype.name string
    Returns:
        type as a C string
    """
    if name == 'float64':
        return 'double'
    elif name == 'float32':
        return 'float'
    elif name.startswith('int'):
        width = int(name[len("int"):])
        return f"int{width}_t"
    elif name.startswith('uint'):
        width = int(name[len("uint"):])
        return f"uint{width}_t"
    elif name == 'bool':
        return 'bool'
    else:
        raise NotImplementedError(f"Can't map numpy to C name for {name}")


class AbstractType(sp.Atom):
    # TODO: Is it necessary to ineherit from sp.Atom?
    def __new__(cls, *args, **kwargs):
        return sp.Basic.__new__(cls)

    def _sympystr(self, *args, **kwargs):
        return str(self)

    @property
    @abstractmethod
    def base_type(self) -> Union[None, 'BasicType']:
        """
        Returns: Returns BasicType of a Vector or Pointer type, None otherwise
        """
        pass

    @property
    @abstractmethod
    def item_size(self) -> int:
        """
        Returns: Number of items.
        E.g. width * item_size(basic_type) in vector's case, or simple numpy itemsize in Struct's case.
        """
        pass


class BasicType(AbstractType):
    """
    BasicType is defined with a const qualifier and a np.dtype.
    """

    def __init__(self, dtype: Union[np.dtype, 'BasicType', str], const: bool = False):
        if isinstance(dtype, BasicType):
            self.numpy_dtype = dtype.numpy_dtype
            self.const = dtype.const
        else:
            self.numpy_dtype = np.dtype(dtype)
            self.const = const
        assert is_supported_type(self.numpy_dtype), f'Type {self.numpy_dtype} is currently not supported!'

    def __getnewargs__(self):
        return self.numpy_dtype, self.const

    def __getnewargs_ex__(self):
        return (self.numpy_dtype, self.const), {}

    @property
    def base_type(self):
        return None

    @property
    def item_size(self):  # TODO: Do we want self.numpy_type.itemsize????
        return 1

    def is_float(self):
        return issubclass(self.numpy_dtype.type, np.floating)

    def is_int(self):
        return issubclass(self.numpy_dtype.type, np.integer)

    def is_uint(self):
        return issubclass(self.numpy_dtype.type, np.unsignedinteger)

    def is_sint(self):
        return issubclass(self.numpy_dtype.type, np.signedinteger)

    def is_bool(self):
        return issubclass(self.numpy_dtype.type, np.bool_)

    def dtype_eq(self, other):
        if not isinstance(other, BasicType):
            return False
        else:
            return self.numpy_dtype == other.numpy_dtype

    @property
    def c_name(self) -> str:
        return numpy_name_to_c(self.numpy_dtype.name)

    def __str__(self):
        return f'{self.c_name}{" const" if self.const else ""}'

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.dtype_eq(other) and self.const == other.const

    def __hash__(self):
        return hash(str(self))


class VectorType(AbstractType):
    """
    VectorType consists of a BasicType and a width.
    """
    instruction_set = None

    def __init__(self, base_type: BasicType, width: int):
        self._base_type = base_type
        self.width = width

    @property
    def base_type(self):
        return self._base_type

    @property
    def item_size(self):
        return self.width * self.base_type.item_size

    def __eq__(self, other):
        if not isinstance(other, VectorType):
            return False
        else:
            return (self.base_type, self.width) == (other.base_type, other.width)

    def __str__(self):
        if self.instruction_set is None:
            return f"{self.base_type}[{self.width}]"
        else:
            # TODO VectorizationRevamp: this seems super weird. the instruction_set should know how to print a type out!
            # TODO VectorizationRevamp: this is error prone. base_type could be cons=True. Use dtype instead
            if self.base_type == create_type("int64") or self.base_type == create_type("int32"):
                return self.instruction_set['int']
            elif self.base_type == create_type("float64"):
                return self.instruction_set['double']
            elif self.base_type == create_type("float32"):
                return self.instruction_set['float']
            elif self.base_type == create_type("bool"):
                return self.instruction_set['bool']
            else:
                raise NotImplementedError()

    def __hash__(self):
        return hash((self.base_type, self.width))

    def __getnewargs__(self):
        return self._base_type, self.width

    def __getnewargs_ex__(self):
        return (self._base_type, self.width), {}


class PointerType(AbstractType):
    def __init__(self, base_type: BasicType, const: bool = False, restrict: bool = True):
        self._base_type = base_type
        self.const = const
        self.restrict = restrict

    def __getnewargs__(self):
        return self.base_type, self.const, self.restrict

    def __getnewargs_ex__(self):
        return (self.base_type, self.const, self.restrict), {}

    @property
    def alias(self):
        return not self.restrict

    @property
    def base_type(self):
        return self._base_type

    @property
    def item_size(self):
        return self.base_type.item_size

    def __eq__(self, other):
        if not isinstance(other, PointerType):
            return False
        else:
            return (self.base_type, self.const, self.restrict) == (other.base_type, other.const, other.restrict)

    def __str__(self):
        return f'{str(self.base_type)} * {"RESTRICT " if self.restrict else "" }{"const" if self.const else ""}'

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash((self._base_type, self.const, self.restrict))


class StructType(AbstractType):
    """
    A list of types (with C offsets).
    It is implemented with uint8_t and casts to the correct datatype.
    """
    def __init__(self, numpy_type, const=False):
        self.const = const
        self._dtype = np.dtype(numpy_type)

    def __getnewargs__(self):
        return self.numpy_dtype, self.const

    def __getnewargs_ex__(self):
        return (self.numpy_dtype, self.const), {}

    @property
    def base_type(self):
        return None

    @property
    def numpy_dtype(self):
        return self._dtype

    @property
    def item_size(self):
        return self.numpy_dtype.itemsize

    def get_element_offset(self, element_name):
        return self.numpy_dtype.fields[element_name][1]

    def get_element_type(self, element_name):
        np_element_type = self.numpy_dtype.fields[element_name][0]
        return BasicType(np_element_type, self.const)

    def has_element(self, element_name):
        return element_name in self.numpy_dtype.fields

    def __eq__(self, other):
        if not isinstance(other, StructType):
            return False
        else:
            return (self.numpy_dtype, self.const) == (other.numpy_dtype, other.const)

    def __str__(self):
        # structs are handled byte-wise
        result = "uint8_t"
        if self.const:
            result += " const"
        return result

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash((self.numpy_dtype, self.const))


def create_type(specification: Union[np.dtype, AbstractType, str]) -> AbstractType:
    # TODO: Deprecated Use the constructor of BasicType or StructType instead
    """Creates a subclass of Type according to a string or an object of subclass Type.

    Args:
        specification: Type object, or a string

    Returns:
        Type object, or a new Type object parsed from the string
    """
    if isinstance(specification, AbstractType):
        return specification
    else:
        numpy_dtype = np.dtype(specification)
        if numpy_dtype.fields is None:
            return BasicType(numpy_dtype, const=False)
        else:
            return StructType(numpy_dtype, const=False)



from pystencils.typing.cast_functions import (CastFunc, BooleanCastFunc, VectorMemoryAccess, ReinterpretCastFunc,
                                              PointerArithmeticFunc)
from pystencils.typing.types import (is_supported_type, numpy_name_to_c, AbstractType, BasicType, VectorType,
                                     PointerType, StructType, create_type)
from pystencils.typing.typed_sympy import (assumptions_from_dtype, TypedSymbol, FieldStrideSymbol, FieldShapeSymbol,
                                           FieldPointerSymbol)
from pystencils.typing.utilities import (typed_symbols, get_base_type, result_type, collate_types,
                                         get_type_of_expression, get_next_parent_of_type, parents_of_type)


__all__ = ['CastFunc', 'BooleanCastFunc', 'VectorMemoryAccess', 'ReinterpretCastFunc', 'PointerArithmeticFunc',
           'is_supported_type', 'numpy_name_to_c', 'AbstractType', 'BasicType',
           'VectorType', 'PointerType', 'StructType', 'create_type',
           'assumptions_from_dtype', 'TypedSymbol', 'FieldStrideSymbol', 'FieldShapeSymbol', 'FieldPointerSymbol',
           'typed_symbols', 'get_base_type', 'result_type', 'collate_types',
           'get_type_of_expression', 'get_next_parent_of_type', 'parents_of_type']



from collections import namedtuple
from typing import Union, Tuple, Any, DefaultDict
import logging

import numpy as np

import sympy as sp
from sympy import Piecewise
from sympy.core.relational import Relational
from sympy.functions.elementary.piecewise import ExprCondPair
from sympy.functions.elementary.trigonometric import TrigonometricFunction, InverseTrigonometricFunction
from sympy.functions.elementary.hyperbolic import HyperbolicFunction
from sympy.codegen import Assignment
from sympy.logic.boolalg import BooleanFunction
from sympy.logic.boolalg import BooleanAtom

from pystencils import astnodes as ast
from pystencils.functions import DivFunc, AddressOf
from pystencils.cpu.vectorization import vec_all, vec_any
from pystencils.field import Field
from pystencils.typing.types import BasicType, PointerType
from pystencils.typing.utilities import collate_types
from pystencils.typing.cast_functions import CastFunc, BooleanCastFunc
from pystencils.typing.typed_sympy import TypedSymbol
from pystencils.fast_approximation import fast_sqrt, fast_division, fast_inv_sqrt
from pystencils.utils import ContextVar


class TypeAdder:
    # TODO: specification -> jupyter notebook
    """Checks if the input to create_kernel is valid.

    Test the following conditions:

    - SSA Form for pure symbols:
        -  Every pure symbol may occur only once as left-hand-side of an assignment
        -  Every pure symbol that is read, may not be written to later
    - Independence / Parallelization condition:
        - a field that is written may only be read at exact the same spatial position

    (Pure symbols are symbols that are not Field.Accesses)
    """
    FieldAndIndex = namedtuple('FieldAndIndex', ['field', 'index'])

    def __init__(self, type_for_symbol: DefaultDict[str, BasicType], default_number_float: BasicType,
                 default_number_int: BasicType):
        self.type_for_symbol = type_for_symbol
        self.default_number_float = ContextVar(default_number_float)
        self.default_number_int = ContextVar(default_number_int)

    def visit(self, obj):
        if isinstance(obj, (list, tuple)):
            return [self.visit(e) for e in obj]
        if isinstance(obj, (sp.Eq, ast.SympyAssignment, Assignment)):
            return self.process_assignment(obj)
        elif isinstance(obj, ast.Conditional):
            condition, condition_type = self.figure_out_type(obj.condition_expr)
            assert condition_type == BasicType('bool')
            true_block = self.visit(obj.true_block)
            false_block = None if obj.false_block is None else self.visit(
                obj.false_block)
            return ast.Conditional(condition, true_block=true_block, false_block=false_block)
        elif isinstance(obj, ast.Block):
            return ast.Block([self.visit(e) for e in obj.args])
        elif isinstance(obj, ast.Node) and not isinstance(obj, ast.LoopOverCoordinate):
            return obj
        else:
            raise ValueError("Invalid object in kernel " + str(type(obj)))

    def process_assignment(self, assignment: Union[sp.Eq, ast.SympyAssignment, Assignment]) -> ast.SympyAssignment:
        # for checks it is crucial to process rhs before lhs to catch e.g. a = a + 1
        new_rhs, rhs_type = self.figure_out_type(assignment.rhs)

        lhs = assignment.lhs
        if not isinstance(lhs, (Field.Access, TypedSymbol)):
            if isinstance(lhs, sp.Symbol):
                self.type_for_symbol[lhs.name] = rhs_type
            else:
                raise ValueError(f'Lhs: `{lhs}` is not a subtype of sp.Symbol')
        new_lhs, lhs_type = self.figure_out_type(lhs)
        assert isinstance(new_lhs, (Field.Access, TypedSymbol))

        if lhs_type != rhs_type:
            logging.warning(f'Lhs"{new_lhs} of type "{lhs_type}" is assigned with a different datatype '
                            f'rhs: "{new_rhs}" of type "{rhs_type}".')
            return ast.SympyAssignment(new_lhs, CastFunc(new_rhs, lhs_type))
        else:
            return ast.SympyAssignment(new_lhs, new_rhs)

    # Type System Specification
    # - Defined Types: TypedSymbol, Field, Field.Access, ...?
    # - Indexed: always unsigned_integer64
    # - Undefined Types: Symbol
    #       - Is specified in Config in the dict or as 'default_type' or behaves like `auto` in the case of lhs.
    # - Constants/Numbers: Are either integer or floating. The precision and sign is specified via config
    #       - Example: 1.4 config:float32 -> float32
    # - Expressions deduce types from arguments
    # - Functions deduce types from arguments
    # - default_type and default_float and default_int can be given for a list of assignment, or
    #   individually as a list for assignment

    # Possible Problems - Do we need to support this?
    # - Mixture in expression with int and float
    # - Mixture in expression with uint64 and sint64
    # TODO Logging: Lowest log level should log all casts ----> cast factory, make cast should contain logging
    def figure_out_type(self, expr) -> Tuple[Any, Union[BasicType, PointerType]]:
        # Trivial cases
        from pystencils.field import Field
        import pystencils.integer_functions
        from pystencils.bit_masks import flag_cond
        bool_type = BasicType('bool')

        # TOOO: check the access
        if isinstance(expr, Field.Access):
            return expr, expr.dtype
        elif isinstance(expr, TypedSymbol):
            return expr, expr.dtype
        elif isinstance(expr, sp.Symbol):
            t = TypedSymbol(expr.name, self.type_for_symbol[expr.name])
            return t, t.dtype
        elif isinstance(expr, np.generic):
            assert False, f'Why do we have a np.generic in rhs???? {expr}'
        elif isinstance(expr, (sp.core.numbers.Infinity, sp.core.numbers.NegativeInfinity)):
            return expr, BasicType('float32')  # see https://en.cppreference.com/w/cpp/numeric/math/INFINITY
        elif isinstance(expr, sp.Number):
            if expr.is_Integer:
                data_type = self.default_number_int.get()
            elif expr.is_Float or expr.is_Rational:
                data_type = self.default_number_float.get()
            else:
                assert False, f'{sp.Number} is neither Float nor Integer'
            return CastFunc(expr, data_type), data_type
        elif isinstance(expr, AddressOf):
            of = expr.args[0]
            # TODO Basically this should do address_of already
            assert isinstance(of, (Field.Access, TypedSymbol, Field))
            return expr, PointerType(of.dtype)
        elif isinstance(expr, BooleanAtom):
            return expr, bool_type
        elif isinstance(expr, Relational):
            # TODO Jan: Code duplication with general case
            args_types = [self.figure_out_type(arg) for arg in expr.args]
            collated_type = collate_types([t for _, t in args_types])
            if isinstance(expr, sp.Equality) and collated_type.is_float():
                logging.warning(f"Using floating point numbers in equality comparison: {expr}")
            new_args = [a if t.dtype_eq(collated_type) else CastFunc(a, collated_type) for a, t in args_types]
            new_eq = expr.func(*new_args)
            return new_eq, bool_type
        elif isinstance(expr, CastFunc):
            new_expr, _ = self.figure_out_type(expr.expr)
            return expr.func(*[new_expr, expr.dtype]), expr.dtype
        elif isinstance(expr, ast.ConditionalFieldAccess):
            access, access_type = self.figure_out_type(expr.access)
            value, value_type = self.figure_out_type(expr.outofbounds_value)
            condition, condition_type = self.figure_out_type(expr.outofbounds_condition)
            assert condition_type == bool_type
            collated_type = collate_types([access_type, value_type])
            if collated_type == access_type:
                new_access = access
            else:
                logging.warning(f"In {expr} the Field Access had to be casted to {collated_type}. This is "
                                f"probably due to a type missmatch of the Field and the value of "
                                f"ConditionalFieldAccess")
                new_access = CastFunc(access, collated_type)

            new_value = value if value_type == collated_type else CastFunc(value, collated_type)
            return expr.func(new_access, condition, new_value), collated_type
        elif isinstance(expr, (vec_any, vec_all)):
            return expr, bool_type
        elif isinstance(expr, BooleanFunction):
            args_types = [self.figure_out_type(a) for a in expr.args]
            new_args = [a if t.dtype_eq(bool_type) else BooleanCastFunc(a, bool_type) for a, t in args_types]
            return expr.func(*new_args), bool_type
        elif type(expr, ) in pystencils.integer_functions.__dict__.values():
            args_types = [self.figure_out_type(a) for a in expr.args]
            collated_type = collate_types([t for _, t in args_types])
            # TODO: should we downcast to integer? If yes then which integer type?
            if not collated_type.is_int():
                raise ValueError(f"Integer functions need to be used with integer types but {collated_type} was given")

            return expr, collated_type
        elif isinstance(expr, flag_cond):
            #   do not process the arguments to the bit shift - they must remain integers
            args_types = [self.figure_out_type(a) for a in (expr.args[i] for i in range(2, len(expr.args)))]
            collated_type = collate_types([t for _, t in args_types])
            new_expressions = [a if t.dtype_eq(collated_type) else CastFunc(a, collated_type) for a, t in args_types]
            return expr.func(expr.args[0], expr.args[1], *new_expressions), collated_type
        # elif isinstance(expr, sp.Mul):
        #    raise NotImplementedError('sp.Mul')
        #    # TODO can we ignore this and move it to general expr handling, i.e. removing Mul? (See todo in backend)
        #    # args_types = [self.figure_out_type(arg) for arg in expr.args if arg not in (-1, 1)]
        elif isinstance(expr, sp.Indexed):
            typed_symbol = expr.base.label
            return expr, typed_symbol.dtype
        elif isinstance(expr, ExprCondPair):
            expr_expr, expr_type = self.figure_out_type(expr.expr)
            condition, condition_type = self.figure_out_type(expr.cond)
            if condition_type != bool_type:
                logging.warning(f'Condition "{condition}" is of type "{condition_type}" and not "bool"')
            return expr.func(expr_expr, condition), expr_type
        elif isinstance(expr, Piecewise):
            args_types = [self.figure_out_type(arg) for arg in expr.args]
            collated_type = collate_types([t for _, t in args_types])
            new_args = []
            for a, t in args_types:
                if t != collated_type:
                    if isinstance(a, ExprCondPair):
                        new_args.append(a.func(CastFunc(a.expr, collated_type), a.cond))
                    else:
                        new_args.append(CastFunc(a, collated_type))
                else:
                    new_args.append(a)
            return expr.func(*new_args) if new_args else expr, collated_type
        elif isinstance(expr, (sp.Pow, sp.exp, InverseTrigonometricFunction, TrigonometricFunction,
                               HyperbolicFunction, sp.log)):
            args_types = [self.figure_out_type(arg) for arg in expr.args]
            collated_type = collate_types([t for _, t in args_types])
            new_args = [a if t.dtype_eq(collated_type) else CastFunc(a, collated_type) for a, t in args_types]
            new_func = expr.func(*new_args) if new_args else expr
            if collated_type == BasicType('float64'):
                return new_func, collated_type
            else:
                return CastFunc(new_func, collated_type), collated_type
        elif isinstance(expr, (fast_sqrt, fast_division, fast_inv_sqrt)):
            args_types = [self.figure_out_type(arg) for arg in expr.args]
            collated_type = BasicType('float32')
            new_args = [a if t.dtype_eq(collated_type) else CastFunc(a, collated_type) for a, t in args_types]
            new_func = expr.func(*new_args) if new_args else expr
            return CastFunc(new_func, collated_type), collated_type
        elif isinstance(expr, (sp.Add, sp.Mul, sp.Abs, sp.Min, sp.Max, DivFunc, sp.UnevaluatedExpr)):
            args_types = [self.figure_out_type(arg) for arg in expr.args]
            collated_type = collate_types([t for _, t in args_types])
            if isinstance(collated_type, PointerType):
                if isinstance(expr, sp.Add):
                    return expr.func(*[a for a, _ in args_types]), collated_type
                else:
                    raise NotImplementedError(f'Pointer Arithmetic is implemented only for Add, not {expr}')
            new_args = [a if t.dtype_eq(collated_type) else CastFunc(a, collated_type) for a, t in args_types]

            if isinstance(expr, (sp.Add, sp.Mul)):
                return expr.func(*new_args, evaluate=False) if new_args else expr, collated_type
            else:
                return expr.func(*new_args) if new_args else expr, collated_type
        else:
            raise NotImplementedError(f'expr {type(expr)}: {expr} unknown to typing')



import numpy as np
import sympy as sp

import pytest

import pystencils as ps
from pystencils import x_staggered_vector, TypedSymbol
from pystencils.enums import Target


class TestStaggeredDiffusion:
    def _run(self, num_neighbors, target=ps.Target.CPU, openmp=False):
        L = (40, 40)
        D = 0.066
        dt = 1
        T = 100

        dh = ps.create_data_handling(L, periodicity=True, default_target=target)

        c = dh.add_array('c', values_per_cell=1)
        j = dh.add_array('j', values_per_cell=num_neighbors, field_type=ps.FieldType.STAGGERED_FLUX)

        x_staggered = - c[-1, 0] + c[0, 0]
        y_staggered = - c[0, -1] + c[0, 0]
        xy_staggered = - c[-1, -1] + c[0, 0]
        xY_staggered = - c[-1, 1] + c[0, 0]

        jj = j.staggered_access
        divergence = -1 * D / (1 + sp.sqrt(2) if j.index_shape[0] == 4 else 1) * \
            sum([jj(d) / sp.Matrix(ps.stencil.direction_string_to_offset(d)).norm() for d in j.staggered_stencil
                 + [ps.stencil.inverse_direction_string(d) for d in j.staggered_stencil]])

        update = [ps.Assignment(c.center, c.center + dt * divergence)]
        flux = [ps.Assignment(j.staggered_access("W"), x_staggered),
                ps.Assignment(j.staggered_access("S"), y_staggered)]
        if j.index_shape[0] == 4:
            flux += [ps.Assignment(j.staggered_access("SW"), xy_staggered),
                     ps.Assignment(j.staggered_access("NW"), xY_staggered)]

        staggered_kernel = ps.create_staggered_kernel(flux, target=dh.default_target, cpu_openmp=openmp).compile()
        div_kernel = ps.create_kernel(update, target=dh.default_target, cpu_openmp=openmp).compile()

        def time_loop(steps):
            sync = dh.synchronization_function([c.name])
            dh.all_to_gpu()
            for i in range(steps):
                sync()
                dh.run_kernel(staggered_kernel)
                dh.run_kernel(div_kernel)
            dh.all_to_cpu()

        def init():
            dh.fill(c.name, np.nan, ghost_layers=True, inner_ghost_layers=True)
            dh.fill(c.name, 0)
            dh.fill(j.name, np.nan, ghost_layers=True, inner_ghost_layers=True)
            dh.cpu_arrays[c.name][L[0] // 2:L[0] // 2 + 2, L[1] // 2:L[1] // 2 + 2] = 1.0

        init()
        time_loop(T)

        reference = np.empty(L)
        for x in range(L[0]):
            for y in range(L[1]):
                r = np.array([x, y]) - L[0] / 2 + 0.5
                reference[x, y] = (4 * np.pi * D * T)**(-dh.dim / 2) * np.exp(-np.dot(r, r) / (4 * D * T)) * (2**dh.dim)

        assert np.abs(dh.gather_array(c.name) - reference).max() < 5e-4

    def test_diffusion_2(self):
        self._run(2)

    def test_diffusion_4(self):
        self._run(4)

    def test_diffusion_openmp(self):
        self._run(4, openmp=True)


def test_staggered_subexpressions():
    dh = ps.create_data_handling((10, 10), periodicity=True, default_target=Target.CPU)
    j = dh.add_array('j', values_per_cell=2, field_type=ps.FieldType.STAGGERED)
    c = sp.symbols("c")
    assignments = [ps.Assignment(j.staggered_access("W"), c),
                   ps.Assignment(c, 1)]
    ps.create_staggered_kernel(assignments, target=dh.default_target).compile()


def test_staggered_loop_cutting():
    pytest.importorskip('islpy')
    dh = ps.create_data_handling((4, 4), periodicity=True, default_target=Target.CPU)
    j = dh.add_array('j', values_per_cell=4, field_type=ps.FieldType.STAGGERED)
    assignments = [ps.Assignment(j.staggered_access("SW"), 1)]
    ast = ps.create_staggered_kernel(assignments, target=dh.default_target)
    assert not ast.atoms(ps.astnodes.Conditional)


def test_staggered_vector():
    dim = 2
    v = x_staggered_vector(dim)
    ctr0 = TypedSymbol('ctr_0', 'int', nonnegative=True)
    ctr1 = TypedSymbol('ctr_1', 'int', nonnegative=True)
    expected_result = sp.Matrix(tuple((ctr0 + 0.5, ctr1 + 0.5)))
    assert v == expected_result


import numpy as np

import pytest

import pystencils as ps
from pystencils.astnodes import SympyAssignment
from pystencils.node_collection import NodeCollection
from pystencils.rng import PhiloxFourFloats, PhiloxTwoDoubles, AESNIFourFloats, AESNITwoDoubles, random_symbol
from pystencils.backends.simd_instruction_sets import get_supported_instruction_sets
from pystencils.cpu.cpujit import get_compiler_config
from pystencils.typing import TypedSymbol
from pystencils.enums import Target

RNGs = {('philox', 'float'): PhiloxFourFloats, ('philox', 'double'): PhiloxTwoDoubles,
        ('aesni', 'float'): AESNIFourFloats, ('aesni', 'double'): AESNITwoDoubles}

instruction_sets = get_supported_instruction_sets()
if get_compiler_config()['os'] == 'windows':
    # skip instruction sets supported by the CPU but not by the compiler
    if 'avx' in instruction_sets and ('/arch:avx2' not in get_compiler_config()['flags'].lower()
                                      and '/arch:avx512' not in get_compiler_config()['flags'].lower()):
        instruction_sets.remove('avx')
    if 'avx512' in instruction_sets and '/arch:avx512' not in get_compiler_config()['flags'].lower():
        instruction_sets.remove('avx512')


@pytest.mark.parametrize('target, rng', ((Target.CPU, 'philox'), (Target.CPU, 'aesni'), (Target.GPU, 'philox')))
@pytest.mark.parametrize('precision', ('float', 'double'))
@pytest.mark.parametrize('dtype', ('float', 'double'))
def test_rng(target, rng, precision, dtype, t=124, offsets=(0, 0), keys=(0, 0), offset_values=None):
    if target == Target.GPU:
        pytest.importorskip('pycuda')
    if instruction_sets and {'neon', 'sve', 'vsx', 'rvv'}.intersection(instruction_sets) and rng == 'aesni':
        pytest.xfail('AES not yet implemented for this architecture')
    if rng == 'aesni' and len(keys) == 2:
        keys *= 2
    if offset_values is None:
        offset_values = offsets

    dh = ps.create_data_handling((2, 2), default_ghost_layers=0, default_target=target)
    f = dh.add_array("f", values_per_cell=4 if precision == 'float' else 2,
                     dtype=np.float32 if dtype == 'float' else np.float64)
    dh.fill(f.name, 42.0)

    rng_node = RNGs[(rng, precision)](dh.dim, offsets=offsets, keys=keys)
    assignments = [rng_node] + [SympyAssignment(f(i), s) for i, s in enumerate(rng_node.result_symbols)]
    kernel = ps.create_kernel(assignments, target=dh.default_target).compile()

    dh.all_to_gpu()
    kwargs = {'time_step': t}
    if offset_values != offsets:
        kwargs.update({k.name: v for k, v in zip(offsets, offset_values)})
    dh.run_kernel(kernel, **kwargs)
    dh.all_to_cpu()
    arr = dh.gather_array(f.name)
    assert np.logical_and(arr <= 1.0, arr >= 0).all()

    if rng == 'philox' and t == 124 and offsets == (0, 0) and keys == (0, 0) and dh.shape == (2, 2):
        int_reference = np.array([[[3576608082, 1252663339, 1987745383, 348040302],
                                   [1032407765, 970978240, 2217005168, 2424826293]],
                                  [[2958765206, 3725192638, 2623672781, 1373196132],
                                   [850605163, 1694561295, 3285694973, 2799652583]]])
    else:
        pytest.importorskip('randomgen')
        if rng == 'aesni':
            from randomgen import AESCounter
            int_reference = np.empty(dh.shape + (4,), dtype=int)
            for x in range(dh.shape[0]):
                for y in range(dh.shape[1]):
                    r = AESCounter(counter=t + (x + offset_values[0]) * 2 ** 32 + (y + offset_values[1]) * 2 ** 64,
                                   key=keys[0] + keys[1] * 2 ** 32 + keys[2] * 2 ** 64 + keys[3] * 2 ** 96,
                                   mode="sequence")
                    a, b = r.random_raw(size=2)
                    int_reference[x, y, :] = [a % 2 ** 32, a // 2 ** 32, b % 2 ** 32, b // 2 ** 32]
        else:
            from randomgen import Philox
            int_reference = np.empty(dh.shape + (4,), dtype=int)
            for x in range(dh.shape[0]):
                for y in range(dh.shape[1]):
                    r = Philox(counter=t + (x + offset_values[0]) * 2 ** 32 + (y + offset_values[1]) * 2 ** 64 - 1,
                               key=keys[0] + keys[1] * 2 ** 32, number=4, width=32, mode="sequence")
                    int_reference[x, y, :] = r.random_raw(size=4)

    if precision == 'float' or dtype == 'float':
        eps = np.finfo(np.float32).eps
    else:
        eps = np.finfo(np.float64).eps
    if rng == 'aesni':  # precision appears to be slightly worse
        eps = max(1e-12, 2 * eps)

    if precision == 'float':
        reference = int_reference * 2. ** -32 + 2. ** -33
    else:
        x = int_reference[:, :, 0::2]
        y = int_reference[:, :, 1::2]
        z = x ^ y << (53 - 32)
        reference = z * 2. ** -53 + 2. ** -54
    assert np.allclose(arr, reference, rtol=0, atol=eps)


@pytest.mark.parametrize('vectorized', (False, True))
@pytest.mark.parametrize('kind', ('value', 'symbol'))
def test_rng_offsets(kind, vectorized):
    if vectorized:
        test = test_rng_vectorized
        if not instruction_sets:
            pytest.skip("cannot detect CPU instruction set")
    else:
        test = test_rng
    if kind == 'value':
        test(instruction_sets[-1] if vectorized else Target.CPU, 'philox', 'float', 'float', t=8,
             offsets=(6, 7), keys=(5, 309))
    elif kind == 'symbol':
        offsets = (TypedSymbol("x0", np.uint32), TypedSymbol("y0", np.uint32))
        test(instruction_sets[-1] if vectorized else Target.GPU, 'philox', 'float', 'float', t=8,
             offsets=offsets, offset_values=(6, 7), keys=(5, 309))


@pytest.mark.parametrize('target', instruction_sets)
@pytest.mark.parametrize('rng', ('philox', 'aesni'))
@pytest.mark.parametrize('precision,dtype', (('float', 'float'), ('double', 'double')))
def test_rng_vectorized(target, rng, precision, dtype, t=130, offsets=(1, 3), keys=(0, 0), offset_values=None):
    if (target in ['neon', 'vsx', 'rvv'] or target.startswith('sve')) and rng == 'aesni':
        pytest.xfail('AES not yet implemented for this architecture')
    cpu_vectorize_info = {'assume_inner_stride_one': True, 'assume_aligned': True, 'instruction_set': target}

    dh = ps.create_data_handling((131, 131), default_ghost_layers=0, default_target=Target.CPU)
    f = dh.add_array("f", values_per_cell=4 if precision == 'float' else 2,
                     dtype=np.float32 if dtype == 'float' else np.float64, alignment=True)
    dh.fill(f.name, 42.0)
    ref = dh.add_array("ref", values_per_cell=4 if precision == 'float' else 2)

    rng_node = RNGs[(rng, precision)](dh.dim, offsets=offsets)
    assignments = [rng_node] + [SympyAssignment(ref(i), s) for i, s in enumerate(rng_node.result_symbols)]
    kernel = ps.create_kernel(assignments, target=dh.default_target).compile()

    kwargs = {'time_step': t}
    if offset_values is not None:
        kwargs.update({k.name: v for k, v in zip(offsets, offset_values)})
    dh.run_kernel(kernel, **kwargs)

    rng_node = RNGs[(rng, precision)](dh.dim, offsets=offsets)
    assignments = [rng_node] + [SympyAssignment(f(i), s) for i, s in enumerate(rng_node.result_symbols)]
    kernel = ps.create_kernel(assignments, target=dh.default_target, cpu_vectorize_info=cpu_vectorize_info).compile()

    dh.run_kernel(kernel, **kwargs)

    ref_data = dh.gather_array(ref.name)
    data = dh.gather_array(f.name)

    assert np.allclose(ref_data, data)


@pytest.mark.parametrize('vectorized', (False, True))
def test_rng_symbol(vectorized):
    """Make sure that the RNG symbol generator generates symbols and that the resulting code compiles"""
    cpu_vectorize_info = None
    if vectorized:
        if not instruction_sets:
            pytest.skip("cannot detect CPU instruction set")
        else:
            cpu_vectorize_info = {'assume_inner_stride_one': True, 'assume_aligned': True,
                                  'instruction_set': instruction_sets[-1]}

    dh = ps.create_data_handling((8, 8), default_ghost_layers=0, default_target=Target.CPU)
    f = dh.add_array("f", values_per_cell=2 * dh.dim, alignment=True)
    nc = NodeCollection([SympyAssignment(f(i), 0) for i in range(f.shape[-1])])
    subexpressions = []
    rng_symbol_gen = random_symbol(subexpressions, dim=dh.dim)
    for i in range(f.shape[-1]):
        nc.all_assignments[i] = SympyAssignment(nc.all_assignments[i].lhs, next(rng_symbol_gen))
    symbols = [a.rhs for a in nc.all_assignments]
    [nc.all_assignments.insert(0, subexpression) for subexpression in subexpressions]
    assert len(symbols) == f.shape[-1] and len(set(symbols)) == f.shape[-1]
    ps.create_kernel(nc, target=dh.default_target, cpu_vectorize_info=cpu_vectorize_info).compile()


@pytest.mark.parametrize('vectorized', (False, True))
def test_staggered(vectorized):
    """Make sure that the RNG counter can be substituted during loop cutting"""

    dh = ps.create_data_handling((8, 8), default_ghost_layers=0, default_target=Target.CPU)
    j = dh.add_array("j", values_per_cell=dh.dim, field_type=ps.FieldType.STAGGERED_FLUX)
    a = ps.AssignmentCollection([ps.Assignment(j.staggered_access(n), 0) for n in j.staggered_stencil])
    rng_symbol_gen = random_symbol(a.subexpressions, dim=dh.dim, rng_node=PhiloxTwoDoubles)
    a.main_assignments[0] = ps.Assignment(a.main_assignments[0].lhs, next(rng_symbol_gen))
    kernel = ps.create_staggered_kernel(a, target=dh.default_target).compile()

    if not vectorized:
        return
    if not instruction_sets:
        pytest.skip("cannot detect CPU instruction set")
    pytest.importorskip('islpy')
    cpu_vectorize_info = {'assume_inner_stride_one': True, 'assume_aligned': False,
                          'instruction_set': instruction_sets[-1]}

    dh.fill(j.name, 867)
    dh.run_kernel(kernel, seed=5, time_step=309)
    ref_data = dh.gather_array(j.name)

    kernel2 = ps.create_staggered_kernel(a, target=dh.default_target, cpu_vectorize_info=cpu_vectorize_info).compile()

    dh.fill(j.name, 867)
    dh.run_kernel(kernel2, seed=5, time_step=309)
    data = dh.gather_array(j.name)

    assert np.allclose(ref_data, data)



import numpy as np
from numpy.testing import assert_array_equal
from pystencils import create_data_handling
from pystencils.slicing import SlicedGetter, make_slice, SlicedGetterDataHandling, shift_slice, slice_intersection


def test_sliced_getter():
    def get_slice(slice_obj=None):
        arr = np.ones((10, 10))
        if slice_obj is None:
            slice_obj = make_slice[:, :]

        return arr[slice_obj]

    sli = SlicedGetter(get_slice)

    test = make_slice[2:-2, 2:-2]
    assert sli[test].shape == (6, 6)


def test_sliced_getter_data_handling():
    domain_shape = (10, 10)

    dh = create_data_handling(domain_size=domain_shape, default_ghost_layers=1)
    dh.add_array("src", values_per_cell=1)
    dh.fill("src", 1.0, ghost_layers=True)

    dh.add_array("dst", values_per_cell=1)
    dh.fill("dst", 0.0, ghost_layers=True)

    sli = SlicedGetterDataHandling(dh, 'dst')
    slice_obj = make_slice[2:-2, 2:-2]
    assert np.sum(sli[slice_obj]) == 0

    sli = SlicedGetterDataHandling(dh, 'src')
    slice_obj = make_slice[2:-2, 2:-2]
    assert np.sum(sli[slice_obj]) == 36


def test_shift_slice():

    sh = shift_slice(make_slice[2:-2, 2:-2], [1, 2])
    assert sh[0] == slice(3, -1, None)
    assert sh[1] == slice(4, 0, None)

    sh = shift_slice(make_slice[2:-2, 2:-2], 1)
    assert sh[0] == slice(3, -1, None)
    assert sh[1] == slice(3, -1, None)

    sh = shift_slice([2, 4], 1)
    assert sh[0] == 3
    assert sh[1] == 5

    sh = shift_slice([2, None], 1)
    assert sh[0] == 3
    assert sh[1] is None

    sh = shift_slice([1.5, 1.5], 1)
    assert sh[0] == 1.5
    assert sh[1] == 1.5


def test_shifted_array_access():
    arr = np.array(range(10))
    
    sh = make_slice[2:5]
    assert_array_equal(arr[sh], [2,3,4])

    sh = shift_slice(sh, 3)
    assert_array_equal(arr[sh], [5,6,7])

    arr = np.array([
        [1, 2, 3],
        [4, 5, 6],
        [7, 8, 9]
    ])

    sh = make_slice[0:2, 0:2]
    assert_array_equal(arr[sh], [[1, 2], [4, 5]])

    sh = shift_slice(sh, (1,1))
    assert_array_equal(arr[sh], [[5, 6], [8, 9]])


def test_slice_intersection():
    sl1 = make_slice[1:10, 1:10]
    sl2 = make_slice[5:15, 5:15]

    intersection = slice_intersection(sl1, sl2)
    assert intersection[0] == slice(5, 10, None)
    assert intersection[1] == slice(5, 10, None)

    sl2 = make_slice[12:15, 12:15]

    intersection = slice_intersection(sl1, sl2)
    assert intersection is None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
import datetime
import sphinx_rtd_theme
import os
import re
import sys

sys.path.insert(0, os.path.abspath('.'))
import pystencils

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'nbsphinx',
    'sphinxcontrib.bibtex',
    'sphinx_autodoc_typehints',
]

add_module_names = False
templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'

copyright = f'{datetime.datetime.now().year}, Martin Bauer'
author = 'Martin Bauer'
# The short X.Y version (including .devXXXX, rcX, b1 suffixes if present)
version = re.sub(r'(\d+\.\d+)\.\d+(.*)', r'\1\2', pystencils.__version__)
version = re.sub(r'(\.dev\d+).*?$', r'\1', version)
# The full version, including alpha/beta/rc tags.
release = pystencils.__version__
language = None
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']
default_role = 'any'
pygments_style = 'sphinx'
todo_include_todos = False

# Options for HTML output

html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_theme = 'sphinx_rtd_theme'
htmlhelp_basename = 'pystencilsdoc'
html_sidebars = {'**': ['globaltoc.html', 'relations.html', 'sourcelink.html', 'searchbox.html']}

# NbSphinx configuration
nbsphinx_execute = 'never'
nbsphinx_codecell_lexer = 'python3'

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'python': ('https://docs.python.org/3.8', None),
                       'numpy': ('https://docs.scipy.org/doc/numpy/', None),
                       'matplotlib': ('https://matplotlib.org/', None),
                       'sympy': ('https://docs.sympy.org/latest/', None),
                       }

autodoc_member_order = 'bysource'
bibtex_bibfiles = ['sphinx/pystencils.bib']

project = 'pystencils'
html_logo = 'img/logo.png'



from pystencils.astnodes import Node
from pystencils.backends.cbackend import CBackend, CustomSympyPrinter, generate_c
from pystencils.enums import Backend
from pystencils.fast_approximation import fast_division, fast_inv_sqrt, fast_sqrt


def generate_cuda(ast_node: Node, signature_only: bool = False, custom_backend=None, with_globals=True) -> str:
    """Prints an abstract syntax tree node as CUDA code.

    Args:
        ast_node: ast representation of kernel
        signature_only: generate signature without function body
        custom_backend: use own custom printer for code generation
        with_globals: enable usage of global variables

    Returns:
        CUDA code for the ast node and its descendants
    """
    return generate_c(ast_node, signature_only, dialect=Backend.CUDA,
                      custom_backend=custom_backend, with_globals=with_globals)


class CudaBackend(CBackend):

    def __init__(self, sympy_printer=None,
                 signature_only=False):
        if not sympy_printer:
            sympy_printer = CudaSympyPrinter()

        super().__init__(sympy_printer, signature_only, dialect=Backend.CUDA)

    def _print_SharedMemoryAllocation(self, node):
        dtype = node.symbol.dtype
        name = self.sympy_printer.doprint(node.symbol.name)
        num_elements = '*'.join([str(s) for s in node.shared_mem.shape])
        code = f"__shared__ {dtype} {name}[{num_elements}];"
        return code

    @staticmethod
    def _print_ThreadBlockSynchronization(_):
        return "__synchtreads();"

    def _print_TextureDeclaration(self, node):
        cond = node.texture.field.dtype.numpy_dtype.itemsize > 4
        return f'texture<{"fp_tex_" if cond else ""}{str(node.texture.field.dtype)}, ' \
               f'cudaTextureType{node.texture.field.spacial_dimensions}D, cudaReadModeElementType> {node.texture};'

    def _print_SkipIteration(self, _):
        return "return;"


class CudaSympyPrinter(CustomSympyPrinter):
    language = "CUDA"

    def __init__(self):
        super(CudaSympyPrinter, self).__init__()

    def _print_Function(self, expr):
        if isinstance(expr, fast_division):
            assert len(expr.args) == 2, f"__fdividef has two arguments, but {len(expr.args)} where given"
            return f"__fdividef({self._print(expr.args[0])}, {self._print(expr.args[1])})"
        elif isinstance(expr, fast_sqrt):
            assert len(expr.args) == 1, f"__fsqrt_rn has one argument, but {len(expr.args)} where given"
            return f"__fsqrt_rn({self._print(expr.args[0])})"
        elif isinstance(expr, fast_inv_sqrt):
            assert len(expr.args) == 1, f"__frsqrt_rn has one argument, but {len(expr.args)} where given"
            return f"__frsqrt_rn({self._print(expr.args[0])})"
        return super()._print_Function(expr)



import pystencils as ps


def test_version_string():
    version = ps.__version__
    print(version)

    numeric_version = [int(x, 10) for x in version.split('.')[0:1]]
    test_version = sum(x * (100 ** i) for i, x in enumerate(numeric_version))

    assert test_version >= 1



import pytest

import numpy as np

import pystencils.config
import sympy as sp

import pystencils as ps
from pystencils.backends.simd_instruction_sets import (get_cacheline_size, get_supported_instruction_sets,
                                                       get_vector_instruction_set)
from . import test_vectorization
from pystencils.enums import Target

supported_instruction_sets = get_supported_instruction_sets() if get_supported_instruction_sets() else []


@pytest.mark.parametrize('instruction_set', supported_instruction_sets)
def test_vectorisation_varying_arch(instruction_set):
    shape = (9, 9, 3)
    arr = np.ones(shape, order='f')

    @ps.kernel
    def update_rule(s):
        f = ps.fields("f(3) : [2D]", f=arr)
        s.tmp0 @= f(0)
        s.tmp1 @= f(1)
        s.tmp2 @= f(2)
        f0, f1, f2 = f(0), f(1), f(2)
        f0 @= 2 * s.tmp0
        f1 @= 2 * s.tmp0
        f2 @= 2 * s.tmp0

    config = pystencils.config.CreateKernelConfig(cpu_vectorize_info={'instruction_set': instruction_set})
    ast = ps.create_kernel(update_rule, config=config)
    kernel = ast.compile()
    kernel(f=arr)
    np.testing.assert_equal(arr, 2)


@pytest.mark.parametrize('dtype', ('float', 'double'))
@pytest.mark.parametrize('instruction_set', supported_instruction_sets)
def test_vectorized_abs(instruction_set, dtype):
    """Some instructions sets have abs, some don't.
       Furthermore, the special treatment of unary minus makes this data type-sensitive too.
    """
    arr = np.ones((2 ** 2 + 2, 2 ** 3 + 2), dtype=np.float64 if dtype == 'double' else np.float32)
    arr[-3:, :] = -1

    f, g = ps.fields(f=arr, g=arr)
    update_rule = [ps.Assignment(g.center(), sp.Abs(f.center()))]

    config = pystencils.config.CreateKernelConfig(cpu_vectorize_info={'instruction_set': instruction_set})
    ast = ps.create_kernel(update_rule, config=config)

    func = ast.compile()
    dst = np.zeros_like(arr)
    func(g=dst, f=arr)
    np.testing.assert_equal(np.sum(dst[1:-1, 1:-1]), 2 ** 2 * 2 ** 3)


@pytest.mark.parametrize('dtype', ('float', 'double'))
@pytest.mark.parametrize('instruction_set', supported_instruction_sets)
def test_strided(instruction_set, dtype):
    type_string = "float64" if dtype == 'double' else "float32"

    f, g = ps.fields(f"f, g : {type_string}[2D]")
    update_rule = [ps.Assignment(g[0, 0], f[0, 0] + f[-1, 0] + f[1, 0] + f[0, 1] + f[0, -1] + 42.0)]
    if 'storeS' not in get_vector_instruction_set(dtype, instruction_set) and instruction_set not in ['avx512',
                                                                                                      'rvv'] and not instruction_set.startswith(
            'sve'):
        with pytest.warns(UserWarning) as warn:
            config = pystencils.config.CreateKernelConfig(cpu_vectorize_info={'instruction_set': instruction_set},
                                                          default_number_float=type_string)
            ast = ps.create_kernel(update_rule, config=config)
            assert 'Could not vectorize loop' in warn[0].message.args[0]
    else:
        with pytest.warns(None) as warn:
            config = pystencils.config.CreateKernelConfig(cpu_vectorize_info={'instruction_set': instruction_set},
                                                          default_number_float=type_string)
            ast = ps.create_kernel(update_rule, config=config)
            assert len(warn) == 0

    # ps.show_code(ast)
    func = ast.compile()
    ref_config = pystencils.config.CreateKernelConfig(default_number_float=type_string)
    ref_func = ps.create_kernel(update_rule, config=ref_config).compile()

    # For some reason other array creations fail on the emulated ppc pipeline
    size = (25, 19)
    arr = np.zeros(size).astype(type_string)
    for i in range(size[0]):
        for j in range(size[1]):
            arr[i, j] = i * j

    dst = np.zeros_like(arr, dtype=type_string)
    ref = np.zeros_like(arr, dtype=type_string)

    func(g=dst, f=arr)
    ref_func(g=ref, f=arr)

    # print("dst: ", dst)
    # print("np array: ", arr)

    np.testing.assert_almost_equal(dst[1:-1, 1:-1], ref[1:-1, 1:-1], 13 if dtype == 'double' else 5)


@pytest.mark.parametrize('dtype', ('float', 'double'))
@pytest.mark.parametrize('instruction_set', supported_instruction_sets)
@pytest.mark.parametrize('gl_field, gl_kernel', [(1, 0), (0, 1), (1, 1)])
def test_alignment_and_correct_ghost_layers(gl_field, gl_kernel, instruction_set, dtype):
    dtype = np.float64 if dtype == 'double' else np.float32

    domain_size = (128, 128)
    dh = ps.create_data_handling(domain_size, periodicity=(True, True), default_target=Target.CPU)
    src = dh.add_array("src", values_per_cell=1, dtype=dtype, ghost_layers=gl_field, alignment=True)
    dh.fill(src.name, 1.0, ghost_layers=True)
    dst = dh.add_array("dst", values_per_cell=1, dtype=dtype, ghost_layers=gl_field, alignment=True)
    dh.fill(dst.name, 1.0, ghost_layers=True)

    update_rule = ps.Assignment(dst[0, 0], src[0, 0])
    opt = {'instruction_set': instruction_set, 'assume_aligned': True,
           'nontemporal': True, 'assume_inner_stride_one': True}
    config = pystencils.config.CreateKernelConfig(target=dh.default_target,
                                                  cpu_vectorize_info=opt, ghost_layers=gl_kernel)
    ast = ps.create_kernel(update_rule, config=config)
    kernel = ast.compile()
    if gl_kernel != gl_field:
        with pytest.raises(ValueError):
            dh.run_kernel(kernel)
    else:
        dh.run_kernel(kernel)


@pytest.mark.parametrize('instruction_set', supported_instruction_sets)
def test_cacheline_size(instruction_set):
    cacheline_size = get_cacheline_size(instruction_set)
    if cacheline_size is None and instruction_set in ['sse', 'avx', 'avx512', 'rvv']:
        pytest.skip()
    instruction_set = get_vector_instruction_set('double', instruction_set)
    vector_size = instruction_set['bytes']
    assert 8 < cacheline_size < 0x100000, "Cache line size is implausible"
    if type(vector_size) is int:
        assert cacheline_size % vector_size == 0, "Cache line size should be multiple of vector size"
    assert cacheline_size & (cacheline_size - 1) == 0, "Cache line size is not a power of 2"


# TODO move to vectorise
@pytest.mark.parametrize('instruction_set',
                         sorted(set(supported_instruction_sets) - {test_vectorization.instruction_set}))
@pytest.mark.parametrize('function',
                         [f for f in test_vectorization.__dict__ if f.startswith('test_') and f not in ['test_hardware_query', 'test_aligned_and_nt_stores']])
def test_vectorization_other(instruction_set, function):
    test_vectorization.__dict__[function](instruction_set)


@pytest.mark.parametrize('dtype', ('float', 'double'))
@pytest.mark.parametrize('instruction_set', supported_instruction_sets)
@pytest.mark.parametrize('field_layout', ('fzyx', 'zyxf'))
def test_square_root(dtype, instruction_set, field_layout):
    config = pystencils.config.CreateKernelConfig(data_type=dtype,
                                                  cpu_vectorize_info={'instruction_set': instruction_set,
                                                       'assume_inner_stride_one': True,
                                                       'assume_aligned': False, 'nontemporal': False})

    src_field = ps.Field.create_generic('pdfs', 2, dtype, index_dimensions=1, layout=field_layout, index_shape=(9,))

    eq = [ps.Assignment(sp.Symbol("xi"), sum(src_field.center_vector)),
          ps.Assignment(sp.Symbol("xi_2"), sp.Symbol("xi") * sp.sqrt(src_field.center))]

    ps.create_kernel(eq, config=config).compile()



import pytest

from pystencils import create_data_handling
from pystencils.alignedarray import *
from pystencils.field import create_numpy_array_with_layout


def is_aligned(arr, alignment, byte_offset=0):
    address = arr.__array_interface__['data'][0]
    rest = (address + byte_offset) % alignment
    if rest:
        print("Alignment rest", rest)
    return rest == 0


@pytest.mark.parametrize("alignment", [8, 8*4, True])
@pytest.mark.parametrize("shape", [17, 16, (16, 16), (17, 17), (18, 18), (19, 19)])
def test_1d_arrays(alignment, shape):
    arrays = [
        aligned_zeros(shape, alignment),
        aligned_ones(shape, alignment),
        aligned_empty(shape, alignment),
    ]
    for arr in arrays:
        assert is_aligned(arr, alignment)


@pytest.mark.parametrize("order", ['C', 'F'])
@pytest.mark.parametrize("alignment", [8, 8*4, True])
@pytest.mark.parametrize("shape", [(16, 16), (17, 17), (18, 18), (19, 19)])
def test_3d_arrays(order, alignment, shape):
    arrays = [
        aligned_zeros(shape, alignment, order=order),
        aligned_ones(shape, alignment, order=order),
        aligned_empty(shape, alignment, order=order),
    ]
    for arr in arrays:
        assert is_aligned(arr, alignment)
        if order == 'C':
            assert is_aligned(arr[1], alignment)
            assert is_aligned(arr[5], alignment)
        else:
            assert is_aligned(arr[..., 1], alignment)
            assert is_aligned(arr[..., 5], alignment)


@pytest.mark.parametrize("parallel", [False, True])
def test_data_handling(parallel):
    for tries in range(16):  # try a few times, since we might get lucky and get randomly a correct alignment
        dh = create_data_handling((6, 7), default_ghost_layers=1, parallel=parallel)
        dh.add_array('test', alignment=8 * 4, values_per_cell=1)
        for b in dh.iterate(ghost_layers=True, inner_ghost_layers=True):
            arr = b['test']
            assert is_aligned(arr[1:, 3:], 8*4)


def test_alignment_of_different_layouts():
    offset = 1
    byte_offset = 8
    for tries in range(16):  # try a few times, since we might get lucky and get randomly a correct alignment
        arr = create_numpy_array_with_layout((3, 4, 5), layout=(0, 1, 2),
                                             alignment=8*4, byte_offset=byte_offset)
        assert is_aligned(arr[offset, ...], 8*4, byte_offset)

        arr = create_numpy_array_with_layout((3, 4, 5), layout=(2, 1, 0),
                                             alignment=8*4, byte_offset=byte_offset)
        assert is_aligned(arr[..., offset], 8*4, byte_offset)

        arr = create_numpy_array_with_layout((3, 4, 5), layout=(2, 0, 1),
                                             alignment=8*4, byte_offset=byte_offset)
        assert is_aligned(arr[:, 0, :], 8*4, byte_offset)



from collections import defaultdict
from functools import partial
from typing import Tuple, Union, Sequence

import numpy as np
import sympy as sp
from sympy.logic.boolalg import Boolean, BooleanFunction

import pystencils
from pystencils.cache import memorycache_if_hashable
from pystencils.typing.types import BasicType, VectorType, PointerType, create_type
from pystencils.typing.cast_functions import CastFunc
from pystencils.typing.typed_sympy import TypedSymbol
from pystencils.utils import all_equal


def typed_symbols(names, dtype, **kwargs):
    """
    Creates TypedSymbols with the same functionality as sympy.symbols
    Args:
        names: See sympy.symbols
        dtype: The data type all symbols will have
        **kwargs: Key value arguments passed to sympy.symbols

    Returns:
        TypedSymbols
    """
    symbols = sp.symbols(names, **kwargs)
    if isinstance(symbols, Tuple):
        return tuple(TypedSymbol(str(s), dtype) for s in symbols)
    else:
        return TypedSymbol(str(symbols), dtype)


def get_base_type(data_type):
    """
    Returns the BasicType of a Pointer or a Vector
    """
    while data_type.base_type is not None:
        data_type = data_type.base_type
    return data_type


def result_type(*args: np.dtype):
    """Returns the type of the result if the np.dtype arguments would be collated.
    We can't use numpy functionality, because numpy casts don't behave exactly like C casts"""
    s = sorted(args, key=lambda x: x.itemsize)

    def kind_to_value(kind: str) -> int:
        if kind == 'f':
            return 3
        elif kind == 'i':
            return 2
        elif kind == 'u':
            return 1
        elif kind == 'b':
            return 0
        else:
            raise NotImplementedError(f'{kind=} is not a supported kind of a type. See "numpy.dtype.kind" for options')
    s = sorted(s, key=lambda x: kind_to_value(x.kind))
    return s[-1]


def collate_types(types: Sequence[Union[BasicType, VectorType]]):
    """
    Takes a sequence of types and returns their "common type" e.g. (float, double, float) -> double
    Uses the collation rules from numpy.
    """
    # Pointer arithmetic case i.e. pointer + [int, uint] is allowed
    if any(isinstance(t, PointerType) for t in types):
        pointer_type = None
        for t in types:
            if isinstance(t, PointerType):
                if pointer_type is not None:
                    raise ValueError(f'Cannot collate the combination of two pointer types "{pointer_type}" and "{t}"')
                pointer_type = t
            elif isinstance(t, BasicType):
                if not (t.is_int() or t.is_uint()):
                    raise ValueError("Invalid pointer arithmetic")
            else:
                raise ValueError("Invalid pointer arithmetic")
        return pointer_type

    # # peel of vector types, if at least one vector type occurred the result will also be the vector type
    vector_type = [t for t in types if isinstance(t, VectorType)]
    if not all_equal(t.width for t in vector_type):
        raise ValueError("Collation failed because of vector types with different width")

    # TODO: check if this is needed
    # def peel_off_type(dtype, type_to_peel_off):
    #     while type(dtype) is type_to_peel_off:
    #         dtype = dtype.base_type
    #     return dtype
    # types = [peel_off_type(t, VectorType) for t in types]

    types = [t.base_type if isinstance(t, VectorType) else t for t in types]

    # now we should have a list of basic types - struct types are not yet supported
    assert all(type(t) is BasicType for t in types)

    result_numpy_type = result_type(*(t.numpy_dtype for t in types))
    result = BasicType(result_numpy_type)
    if vector_type:
        result = VectorType(result, vector_type[0].width)
    return result


# TODO get_type_of_expression should be used after leaf_typing. So no defaults should be necessary
@memorycache_if_hashable(maxsize=2048)
def get_type_of_expression(expr,
                           default_float_type='double',
                           default_int_type='int',
                           symbol_type_dict=None):
    from pystencils.astnodes import ResolvedFieldAccess
    from pystencils.cpu.vectorization import vec_all, vec_any

    if default_float_type == 'float':
        default_float_type = 'float32'

    if not symbol_type_dict:
        symbol_type_dict = defaultdict(lambda: create_type('double'))

    # TODO this line is quite hard to understand, if possible simpl
    get_type = partial(get_type_of_expression,
                       default_float_type=default_float_type,
                       default_int_type=default_int_type,
                       symbol_type_dict=symbol_type_dict)

    expr = sp.sympify(expr)
    if isinstance(expr, sp.Integer):
        return create_type(default_int_type)
    elif isinstance(expr, sp.Rational) or isinstance(expr, sp.Float):
        return create_type(default_float_type)
    elif isinstance(expr, ResolvedFieldAccess):
        return expr.field.dtype
    elif isinstance(expr, pystencils.field.Field.Access):
        return expr.field.dtype
    elif isinstance(expr, TypedSymbol):
        return expr.dtype
    elif isinstance(expr, sp.Symbol):
        # TODO delete if case
        if symbol_type_dict:
            return symbol_type_dict[expr.name]
        else:
            raise ValueError("All symbols inside this expression have to be typed! ", str(expr))
    elif isinstance(expr, CastFunc):
        return expr.args[1]
    elif isinstance(expr, (vec_any, vec_all)):
        return create_type("bool")
    elif hasattr(expr, 'func') and expr.func == sp.Piecewise:
        collated_result_type = collate_types(tuple(get_type(a[0]) for a in expr.args))
        collated_condition_type = collate_types(tuple(get_type(a[1]) for a in expr.args))
        if type(collated_condition_type) is VectorType and type(collated_result_type) is not VectorType:
            collated_result_type = VectorType(collated_result_type, width=collated_condition_type.width)
        return collated_result_type
    elif isinstance(expr, sp.Indexed):
        typed_symbol = expr.base.label
        return typed_symbol.dtype.base_type
    elif isinstance(expr, (Boolean, BooleanFunction)):
        # if any arg is of vector type return a vector boolean, else return a normal scalar boolean
        result = create_type("bool")
        vec_args = [get_type(a) for a in expr.args if isinstance(get_type(a), VectorType)]
        if vec_args:
            result = VectorType(result, width=vec_args[0].width)
        return result
    elif isinstance(expr, sp.Pow):
        base_type = get_type(expr.args[0])
        if expr.exp.is_integer:
            return base_type
        else:
            return collate_types([create_type(default_float_type), base_type])
    elif isinstance(expr, (sp.Sum, sp.Product)):
        return get_type(expr.args[0])
    elif isinstance(expr, sp.Expr):
        expr: sp.Expr
        if expr.args:
            types = tuple(get_type(a) for a in expr.args)
            return collate_types(types)
        else:
            if expr.is_integer:
                return create_type(default_int_type)
            else:
                return create_type(default_float_type)

    raise NotImplementedError("Could not determine type for", expr, type(expr))


# Fix for sympy versions from 1.9
sympy_version = sp.__version__.split('.')
if int(sympy_version[0]) * 100 + int(sympy_version[1]) >= 109:
    # __setstate__ would bypass the contructor, so we remove it
    sp.Number.__getstate__ = sp.Basic.__getstate__
    del sp.Basic.__getstate__

    class FunctorWithStoredKwargs:
        def __init__(self, func, **kwargs):
            self.func = func
            self.kwargs = kwargs

        def __call__(self, *args):
            return self.func(*args, **self.kwargs)

    # __reduce_ex__ would strip kwargs, so we override it
    def basic_reduce_ex(self, protocol):
        if hasattr(self, '__getnewargs_ex__'):
            args, kwargs = self.__getnewargs_ex__()
        else:
            args, kwargs = self.__getnewargs__(), {}
        if hasattr(self, '__getstate__'):
            state = self.__getstate__()
        else:
            state = None
        return FunctorWithStoredKwargs(type(self), **kwargs), args, state

    sp.Number.__reduce_ex__ = sp.Basic.__reduce_ex__
    sp.Basic.__reduce_ex__ = basic_reduce_ex


def get_next_parent_of_type(node, parent_type):
    """Returns the next parent node of given type or None, if root is reached.

    Traverses the AST nodes parents until a parent of given type was found.
    If no such parent is found, None is returned
    """
    parent = node.parent
    while parent is not None:
        if isinstance(parent, parent_type):
            return parent
        parent = parent.parent
    return None


def parents_of_type(node, parent_type, include_current=False):
    """Generator for all parent nodes of given type"""
    parent = node if include_current else node.parent
    while parent is not None:
        if isinstance(parent, parent_type):
            yield parent
        parent = parent.parent
